import json
import itertools
from pathlib import Path

import numpy as np
from scipy.stats import norm
from scipy.optimize import linear_sum_assignment as optimise
from lmfit import Parameters

from dp5.nmr_processing.proton.simulate import new_first_order_peak


def solvent_removal(solvent, y_data, x_data, picked_peaks, peak_regions, grouped_peaks, total_params, uc):
    picked_peaks = np.array(picked_peaks)

    solvent_file = (Path(__file__).parent / 'solvents.json').resolve()

    with open(solvent_file) as f:
        solvent_dict = json.load(f)

    if solvent in solvent_dict:
        solvent_data = solvent_dict[solvent]
    else:
        solvent_data = []

    picked_peaks_ppm = x_data[picked_peaks]

    # make differences vector for referencing against multiple solvent peaks
    differences = []
    peaks_to_remove = []
    solvent_regions = []

    for peak in solvent_data:
        speak_ppm = peak["exp_ppm"]
        # if only a singlet is expected for this peak find solvent peak based on amplitude and position
        if not peak["Jv"]:
            probs = norm.pdf(abs(picked_peaks_ppm - speak_ppm),
                             loc=0, scale=0.1) * y_data[picked_peaks]

            # find the maximum probability
            w = np.argmax(probs)

            # append this to the list to remove
            peaks_to_remove.append(picked_peaks[w])

            # append this to the list of differences
            differences.append(speak_ppm - picked_peaks_ppm[w])
        else:
            amp_res = []
            dist_res = []
            pos_res = []

            # limit the search to peaks +- 1 ppm either side

            srange = (picked_peaks_ppm > speak_ppm - 1) * \
                (picked_peaks_ppm < speak_ppm + 1)

            for peak in picked_peaks_ppm[srange]:
                fit_s_peaks, amp_vector, fit_s_y = new_first_order_peak(peak, peak["Jv"], np.arange(len(x_data)), 0.1, uc,
                                                                        1)

                diff_matrix = np.zeros((len(fit_s_peaks), len(picked_peaks)))

                for i, f in enumerate(fit_s_peaks):

                    for j, g in enumerate(picked_peaks):
                        diff_matrix[i, j] = abs(f - g)

                # minimise these distances
                vertical_ind, horizontal_ind = optimise(diff_matrix)
                closest_peaks = np.sort(picked_peaks[horizontal_ind])
                closest_amps = []

                for cpeak in closest_peaks:
                    closest_amps.append(total_params['A' + str(cpeak)])

                # find the amplitude residual between the closest peaks and the predicted pattern
                # normalise these amplitudes
                amp_vector = [i / max(amp_vector) for i in amp_vector]
                closest_amps = [i / max(closest_amps) for i in closest_amps]

                # append to the vector
                amp_res.append(
                    sum([abs(amp_vector[i] - closest_amps[i]) for i in range(len(amp_vector))]))
                dist_res.append(np.sum(np.abs(closest_peaks - fit_s_peaks)))
                pos_res.append(
                    norm.pdf(abs(peak - speak_ppm), loc=0, scale=0.5))

            # use the gsd data to find amplitudes of these peaks
            pos_res = [1 - i / max(pos_res) for i in pos_res]
            dist_res = [i / max(dist_res) for i in dist_res]
            amp_res = [i / max(amp_res) for i in amp_res]

            # calculate geometric mean of metrics for each peak
            g_mean = [(dist_res[i] + amp_res[i] + pos_res[i]) /
                      3 for i in range(0, len(amp_res))]

            # compare the residuals and find the minimum
            minres = np.argmin(g_mean)

            # append the closest peaks to the vector
            fit_s_peaks, amp_vector, fit_s_y = new_first_order_peak(picked_peaks_ppm[srange][minres], peak["Jv"],
                                                                    np.arange(len(x_data)), 0.1, uc, 1)
            diff_matrix = np.zeros((len(fit_s_peaks), len(picked_peaks)))

            for i, f in enumerate(fit_s_peaks):
                for j, g in enumerate(picked_peaks):
                    diff_matrix[i, j] = abs(f - g)

            # minimise these distances
            vertical_ind, horizontal_ind = optimise(diff_matrix)
            closest_peaks = np.sort(picked_peaks[horizontal_ind])

            for peak in closest_peaks:
                ind3 = np.abs(picked_peaks - peak).argmin()

                peaks_to_remove.append(picked_peaks[ind3])

                differences.append(picked_peaks_ppm[ind3] - uc.ppm(peak))

    # find the region this peak is in and append it to the list
    for peak in peaks_to_remove:
        for ind2, region in enumerate(peak_regions):
            if (peak > region[0]) & (peak < region[-1]):
                solvent_regions.append(ind2)
                break

    # now remove the selected peaks from the picked peaks list and grouped peaks

    w = np.searchsorted(picked_peaks, peaks_to_remove)
    picked_peaks = np.delete(picked_peaks, w)

    for ind4, peak in enumerate(peaks_to_remove):
        grouped_peaks[solvent_regions[ind4]] = np.delete(grouped_peaks[solvent_regions[ind4]],
                                                         np.where(grouped_peaks[solvent_regions[ind4]] == peak)).tolist()

    # resimulate the solvent regions
    solvent_region_ind = sorted(list(set(solvent_regions)))

    # now need to reference the spectrum

    # differences = list of differences in ppm found_solvent_peaks - expected_solvent_peaks

    s_differences = sum(differences)
    x_data = x_data + s_differences

    return peak_regions, picked_peaks, grouped_peaks, x_data, solvent_region_ind



"""Runs DP5 analysis for organic molecules."""

from pathlib import Path
import logging
import dill as pickle
from abc import abstractmethod

from tqdm import tqdm
from joblib import Parallel, delayed
from scipy.stats.kde import gaussian_kde
from scipy.stats import norm
from scipy.optimize import curve_fit
from sklearn.neighbors import KernelDensity

from dp5.neural_net.CNN_model import *
from dp5.analysis.utils import scale_nmr, AnalysisData

logger = logging.getLogger(__name__)


class DP5:
    """Performs DP5 analysis"""

    def __init__(self, output_folder: Path, use_dft_shifts: bool):
        """Initialise the settings.

        Arguments:
          output_folder(Path): path for saved DP5 data
          use_dft_shifts(bool): if set, analyses errors of DFT calculations, compares shifts againt their environments otherwise.

        """
        logger.info("Setting up DP5 method")
        self.output_folder = output_folder
        self.dft_shifts = use_dft_shifts

        if use_dft_shifts:
            # must load model for error prediction
            self.C_DP5 = ErrorDP5ProbabilityCalculator(
                atom_type="C",
                model_file="NMRdb-CASCADEset_Error_mean_model_atom_features256.hdf5",
                batch_size=16,
                transform_file="pca_10_ERRORrep_Error_decomp.p",
                kde_file="pca_10_kde_ERRORrep_Error_kernel.p",
                dp5_correct_scaling="Error_correct_kde.p",
                dp5_incorrect_scaling="Error_incorrect_kde.p",
            )
        else:
            # must load model for shift preiction
            self.C_DP5 = QuantileDP5ProbabilityCalculator(
                atom_type="C",
                model_file="NMRdb_CASCADE_99quantiles.zip",
                batch_size=16,
            )

        if not self.output_folder.exists():
            self.output_folder.mkdir()

        if not (self.output_folder / "dp5").exists():
            (self.output_folder / "dp5").mkdir()

    def __call__(self, mols):
        """Runs DP5 calculations.

        Arguments:
          mols: Molecule objects
        """
        data_dic_path = self.output_folder / "dp5" / "data_dic.p"
        dp5_data = DP5Data(mols, data_dic_path)
        if dp5_data.exists:
            logger.info("Found existing DP5 probability file")
            dp5_data.load()
        else:
            logger.info("Calculating DP5 probabilites...")
            (
                dp5_data.Clabels,
                dp5_data.Cshifts,
                dp5_data.Cexp,
                dp5_data.Cerrors,
                dp5_data.Cconf_atom_probs,
                dp5_data.CDP5_atom_probs,
                dp5_data.CDP5_mol_probs,
            ) = self.C_DP5(mols)
            dp5_data.save()
        return dp5_data.output


class DP5ProbabilityCalculator:
    def __init__(
        self,
        atom_type,
    ):
        """Initialises DP5 Probability calculator for one atom.

        Arguments:
          atom_type (str): atom symbol. Can be 'C' or 'H'
          model_file (str): path for representation generating model to load.
          batch_size (int): batch size for the model.
          transform_file (str): Path to the Scikit-learn PCA file relative to ``dp5/analysis`` folder. Reduces dimensionality of the representation.
          kde_file (str): Path to :py:obj:`scipy.stats.gaussian_kde` or :py:obj:`sklearn.neighbors.KernelDensity` object. Estimates DP5 probabilities
          dp5_correct_scaling (str). Path to :py:obj:`scipy.stats.gaussian_kde` or :py:obj:`sklearn.neighbors.KernelDensity` object. Estimates :math:`P(correct|structure)` for rescaling. Default is None (no scaling)
          dp5_incorrect_scaling (str). Path to :py:obj:`scipy.stats.gaussian_kde` or :py:obj:`sklearn.neighbors.KernelDensity` object. Estimates :math:`P(incorrect|structure)` for rescaling. Default is None (no scaling)

        """
        self.atom_type = atom_type

    @abstractmethod
    def rescale_probabilities(self, mol_probs, errors, error_threshold=2):
        """
        Scales and aggregated atomic probabilities.

        Computes geometric means of atomic probabilities to generate final molecular probabilities.
        """
        total_probs = np.array([np.exp(np.log(arr + 1e-6).mean()) for arr in mol_probs])
        return mol_probs, total_probs

    @abstractmethod
    def probfunction(self, df):
        return NotImplementedError("KDE sampling function not implemented")

    @staticmethod
    def boltzmann_weight(df, col):
        return df.groupby("mol_id")[["conf_population", col]].apply(
            lambda x: (x[col] * x["conf_population"]).sum()
        )

    def __call__(self, mols):
        """Carries out DP5 analysis.

        Arguments:
          mols(list of :py:class:`~dp5.run.data_structures.Molecule`): :py:class:`dp5.run.data_structures.Molecule` objects used in the calculation. Must contain shifts and labels for the provided atom.

        Returns:
          A tuple containing lists of labels of atoms used in the analysis,
          their calculated shifts, their experimental shifts,
          scaled errors, DP5 probabilities for each atom in each conformer,
          Boltzmann-weighted atom DP5 probabilites, and total molecular DP5 probabilities.
        """
        all_labels = []
        rep_df = []
        for mol_id, mol in enumerate(mols):
            calculated, experimental, labels, indices = self.get_shifts_and_labels(mol)
            # drop unassigned !
            has_exp = np.isfinite(experimental)
            new_calcs = calculated[:, has_exp]
            new_exps = experimental[has_exp]
            new_labs = labels[has_exp]
            new_inds = indices[has_exp]

            # generate scaled errors
            scaled = scale_nmr(new_calcs, new_exps)
            corrected_errors = scaled - new_exps[np.newaxis, :]

            all_labels.append(new_labs)

            rep_df.append(
                (
                    mol_id,
                    range(mol.conformers.shape[0]),
                    mol.rdkit_mols,
                    new_inds,
                    mol.populations,
                    new_calcs,
                    new_exps,
                    corrected_errors,
                )
            )

        rep_df = pd.DataFrame(
            rep_df,
            columns=[
                "mol_id",
                "conf_id",
                "Mol",
                "atom_index",
                "conf_population",
                "conf_shifts",
                "exp_shifts",
                "errors",
            ],
        )
        # each row of dataframe represents a geometry
        rep_df = rep_df.explode(
            ["conf_id", "Mol", "conf_shifts", "conf_population", "errors"],
            ignore_index=True,
        )
        logger.info("Extracting atomic representations")
        # now return condensed representations! These are now grouped by conformer
        # need to redo to accommodate

        atom_probs = self.probfunction(rep_df)
        # should be abstracted into KDE-based calculator

        weighted_probs = self.boltzmann_weight(rep_df, "atom_probs")
        weighted_probs = 1 - weighted_probs

        weighted_errors = self.boltzmann_weight(rep_df, "errors")
        cmae = weighted_errors.apply(lambda x: np.mean(np.abs(x)))

        # rescale and aggregate probabilities
        weighted_probs, total_probs = self.rescale_probabilities(weighted_probs, cmae)

        calc_shifts_analysed = self.boltzmann_weight(rep_df, "conf_shifts")
        exp_shifts_analysed = rep_df.groupby("mol_id")["exp_shifts"].first()

        # eventually return atomic probs, weighted atomic probs, DP5 scores
        logger.info("Atomic probabilities estimated")
        return (
            all_labels,
            calc_shifts_analysed,
            exp_shifts_analysed,
            weighted_errors,
            atom_probs,
            weighted_probs,
            total_probs,
        )

    def get_shifts_and_labels(self, mol):
        """
        Returns calculated and experimental shifts for nuclei in the molecule.

        Arguments:
          self.atom_type(str): nuclei being analysed
          mols(:py:class:`~dp5.run.data_structures.Molecule`): :py:class:`dp5.run.data_structures.Molecule`. Must contain shifts and labels for the provided atom.

        Returns:
          calculated conformer shifts
          assigned experimental shifts
          0-based indices of relevant atoms
        """
        at = self.atom_type
        conformer_shifts = getattr(mol, "conformer_%s_pred" % at)
        assigned_shifts = getattr(mol, "%s_exp" % at)
        atom_labels = getattr(mol, "%s_labels" % at)
        atom_indices = np.array([int(label[len(at) :]) - 1 for label in atom_labels])

        return conformer_shifts, assigned_shifts, atom_labels, atom_indices


class ErrorDP5ProbabilityCalculator(DP5ProbabilityCalculator):
    def __init__(
        self,
        atom_type,
        model_file,
        batch_size,
        transform_file,
        kde_file,
        dp5_correct_scaling=None,
        dp5_incorrect_scaling=None,
    ):
        super().__init__(atom_type)
        self.model = build_model(model_file=model_file)
        self.batch_size = batch_size
        self.transform = _load_pickle(transform_file)
        self.kde = KernelDensityEstimator(kde_file)
        if dp5_correct_scaling is not None:
            self.dp5_correct_kde = KernelDensityEstimator(dp5_correct_scaling)
        if dp5_incorrect_scaling is not None:
            self.dp5_incorrect_kde = KernelDensityEstimator(dp5_incorrect_scaling)

    def probfunction(self, rep_df):
        logger.debug("Transforming representations")
        rep_df["representations"] = extract_representations(
            self.model, rep_df, self.batch_size
        )
        rep_df["representations"] = rep_df["representations"].apply(
            self.transform.transform
        )
        # should be abstracted into KDE-based calculator
        logger.info("Estimating atomic probabilities")
        rep_df["atom_probs"] = self.kde_probfunction(rep_df)
        atom_probs = [np.stack(df) for i, df in rep_df.groupby("mol_id")["atom_probs"]]
        return atom_probs

    def kde_probfunction(self, df):
        # loop through atoms in the test molecule - generate kde for all of them.
        # implement joblib parallel search
        # check if this has been calculated

        min_value = -20
        max_value = 20
        n_points = 250

        x = np.linspace(min_value, max_value, n_points)

        probs = []
        with Parallel(prefer="threads", n_jobs=-1) as pool:
            for i, (rep, errors) in tqdm(
                df[["representations", "errors"]].iterrows(),
                total=len(df),
                desc="Computing error KDEs",
                leave=True,
            ):
                num_atoms, num_components = rep.shape
                rep_b = np.broadcast_to(
                    rep[:, :, np.newaxis], shape=(num_atoms, num_components, n_points)
                )
                x_b = np.broadcast_to(
                    x[np.newaxis, np.newaxis, :], shape=(num_atoms, 1, n_points)
                )
                point = np.concatenate((rep_b, x_b), axis=1)

                results = pool(delayed(self.kde)(atom) for atom in point[:])

                conf_probs = []
                for pdf, error in zip(results, errors):
                    integral = 0
                    if pdf.sum() != 0:
                        max_x = x[np.argmax(pdf)]

                        low_point = max(min_value, max_x - abs(max_x - error))
                        high_point = min(max_value, max_x + abs(max_x - error))

                        low_bound = np.argmin(np.abs(x - low_point))
                        high_bound = np.argmin(np.abs(x - high_point))

                        bound_integral = np.sum(
                            pdf[min(low_bound, high_bound) : max(low_bound, high_bound)]
                        )
                        integral = bound_integral / pdf.sum()
                    conf_probs.append(integral)
                probs.append(np.array(conf_probs))

        return probs

    def rescale_probabilities(self, mol_probs, errors, error_threshold=2):
        _, total_probs = super().rescale_probabilities(mol_probs, errors)
        scaled_probs = []
        scaled_total = []
        for prob, error, total in zip(mol_probs, errors, total_probs):
            if error < error_threshold:
                vector = np.concatenate((prob, np.atleast_1d(total)))
                correct = self.dp5_correct_kde(vector)
                incorrect = self.dp5_incorrect_kde(vector)
                scaled = correct / (correct + incorrect)
                scaled_probs.append(scaled[:-1])
                scaled_total.append(scaled[-1])
            else:
                scaled_probs.append(prob)
                scaled_total.append(total)
        return scaled_probs, scaled_total


class ExpDP5ProbabilityCalculator(DP5ProbabilityCalculator):
    def __init__(
        self,
        atom_type,
        model_file,
        batch_size,
        transform_file,
        kde_file,
        dp5_correct_scaling=None,
        dp5_incorrect_scaling=None,
    ):
        super().__init__(atom_type)
        self.model = build_model(model_file=model_file)
        self.batch_size = batch_size
        self.transform = _load_pickle(transform_file)
        self.kde = KernelDensityEstimator(kde_file)
        if dp5_correct_scaling is not None:
            self.dp5_correct_kde = KernelDensityEstimator(dp5_correct_scaling)
        if dp5_incorrect_scaling is not None:
            self.dp5_incorrect_kde = KernelDensityEstimator(dp5_incorrect_scaling)

    def probfunction(self, rep_df):
        logger.debug("Transforming representations")
        rep_df["representations"] = extract_representations(
            self.model, rep_df, self.batch_size
        )
        rep_df["representations"] = rep_df["representations"].apply(
            self.transform.transform
        )
        # should be abstracted into KDE-based calculator
        logger.info("Estimating atomic probabilities")
        rep_df["atom_probs"] = self.kde_probfunction(rep_df)
        atom_probs = [np.stack(df) for i, df in rep_df.groupby("mol_id")["atom_probs"]]
        return atom_probs

    def kde_probfunction(self, df):
        """Since the result is compared to the experimental shifts, weights the representations and runs KDE on those."""
        # loop through atoms in the test molecule - generate kde for all of them.
        total_reps = self.boltzmann_weight(df, "representations")
        exp_data = df.groupby("mol_id")["exp_shifts"].first()
        mol_df = pd.DataFrame({"representations": total_reps, "exp_shifts": exp_data})

        min_value = 0
        max_value = 250
        n_points = 250

        x = np.linspace(min_value, max_value, n_points)

        mol_probs = []
        with Parallel(prefer="threads", n_jobs=-1) as pool:
            for i, (rep, exp) in tqdm(
                mol_df[["representations", "exp_shifts"]].iterrows(),
                total=len(mol_df),
                desc="Computing experimental KDEs",
                leave=True,
            ):
                num_atoms, num_components = rep.shape
                rep_b = np.broadcast_to(
                    rep[:, :, np.newaxis], shape=(num_atoms, num_components, n_points)
                )
                x_b = np.broadcast_to(
                    x[np.newaxis, np.newaxis, :], shape=(num_atoms, 1, n_points)
                )
                point = np.concatenate((rep_b, x_b), axis=1)

                results = pool(delayed(self.kde)(atom) for atom in point[:])

                conf_probs = []
                for pdf, value in zip(results, exp):
                    integral = 0
                    if pdf.sum() != 0:
                        max_x = x[np.argmax(pdf)]

                        low_point = max(min_value, max_x - abs(max_x - value))
                        high_point = min(max_value, max_x + abs(max_x - value))

                        low_bound = np.argmin(np.abs(x - low_point))
                        high_bound = np.argmin(np.abs(x - high_point))

                        bound_integral = np.sum(
                            pdf[min(low_bound, high_bound) : max(low_bound, high_bound)]
                        )
                        integral = bound_integral / pdf.sum()
                    conf_probs.append(integral)
                mol_probs.append(np.array(conf_probs))
        consistency_hack = {i: probs for i, probs in enumerate(mol_probs)}
        consistent_probs = df["mol_id"].map(consistency_hack)
        return consistent_probs

    def rescale_probabilities(self, *args, **kwargs):
        return super().rescale_probabilities(*args, **kwargs)


class QuantileDP5ProbabilityCalculator(DP5ProbabilityCalculator):
    def __init__(
        self, atom_type, model_file, batch_size, quantile_regressor="quantile99.zip"
    ):
        super().__init__(atom_type)
        default_path = str(Path(__file__).parent.parent / "neural_net" / model_file)
        self.model = CASCADE_Quantile.load(default_path)
        self.batch_size = batch_size

    def probfunction(self, df):
        # take representations
        df["quantiles"] = extract_representations(self.model, df, self.batch_size)
        df[["mu", "sigma"]] = self.generate_distributions(df["quantiles"])
        atom_probs_all = []
        for i, (mus, sigmas, exps) in df[["mu", "sigma", "exp_shifts"]].iterrows():
            atom_probs = []
            for mu, sigma, peak in zip(mus, sigmas, exps):
                perc = norm.cdf(peak, mu, sigma)
                prob = np.abs(1 - 2 * perc)
                atom_probs.append(prob)
            atom_probs = np.array(atom_probs)
            atom_probs_all.append(atom_probs)
        df["atom_probs"] = atom_probs_all
        atom_probs = [np.stack(df) for i, df in df.groupby("mol_id")["atom_probs"]]
        return atom_probs

    def generate_distributions(self, quantile_col):
        # in principle, should be able to explode then reassemble
        # makes for easier handling
        # that or compute lengths
        atom_nums = quantile_col.apply(len).values
        slice_inds = np.cumsum(atom_nums)[:-1]
        quantiles = np.concatenate(quantile_col.values)

        # fit distributions, then spit out mean and std
        mus = []
        sigmas = []
        for q in quantiles:
            percentiles = self.model.quantiles
            dims = len(percentiles)
            median = q[dims // 2]
            std = (q[dims * 2 // 3] - q[dims // 3]) / 2

            mu, sigma = curve_fit(norm.cdf, q, percentiles, p0=[median, std])[0]
            mus.append(mu)
            sigmas.append(sigma)

        mus = np.array(mus)
        sigmas = np.array(sigmas)
        result_df = pd.DataFrame(
            {
                "mu": np.split(mus, slice_inds),
                "sigma": np.split(sigmas, slice_inds),
            },
            index=quantile_col.index,
        )

        return result_df


class KernelDensityEstimator:
    def __init__(self, path_to_pickle):
        self.estimator = _load_pickle(path_to_pickle)
        if type(self.estimator) is gaussian_kde:
            self.evaluate = self._scipy_estimator
        elif type(self.estimator) is KernelDensity:
            self.evaluate = self._sklearn_estimator

    def __call__(self, data):
        return self.evaluate(data)

    def _scipy_estimator(self, data):
        return self.estimator(data)

    def _sklearn_estimator(self, data):
        return self.estimator.score_samples(data.T)


class DP5Data(AnalysisData):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def output(self):
        """Uncomment when H-DP5 is implemented"""
        output_dict = dict()
        output_dict["C_output"] = []
        # output_dict["H_output"] = []
        output_dict["CDP5_output"] = []
        # output_dict["HDP5_output"] = []
        # output_dict["DP5_output"] = []
        for mol, clab, cshift, cexp, cerr, cpr in zip(
            self.mols,
            self.Clabels,
            self.Cshifts,
            self.Cexp,
            self.Cerrors,
            self.CDP5_atom_probs,
        ):
            output = f"\nAssigned C NMR shift for {mol}:"
            output += self.print_assignment(clab, cshift, cexp, cerr, cpr)
            output_dict["C_output"].append(output)

        # for mol, hlab, hshift, hscal, hexp, herr in zip(
        #    self.mols, self.Hlabels, self.Hshifts, self.Hscaled, self.Hexp, self.Herrors
        # ):
        #    output = f"\nAssigned H NMR shift for {mol}:"
        #    output += self.print_assignment(hlab, hshift, hscal, hexp, herr)
        #    output_dict["H_output"].append(output)

        for mol, cdp5 in zip(self.mols, self.CDP5_mol_probs):
            output_dict["CDP5_output"].append(
                f"Carbon DP5 probability for {mol}: {cdp5}"
            )
        t_dic = [
            dict(zip(output_dict.keys(), values))
            for values in zip(*output_dict.values())
        ]
        dp5_output = "\n\n".join([mol["C_output"] for mol in t_dic])
        dp5_output += "\n\n"
        dp5_output += "\n".join([mol["CDP5_output"] for mol in t_dic])
        return dp5_output

    @staticmethod
    def print_assignment(labels, calculated, exp, error, probs):
        """Prints table for molecule"""

        s = np.argsort(calculated)
        svalues = calculated[s]
        slabels = labels[s]
        sexp = exp[s]
        serror = error[s]
        sprob = probs[s]

        output = f"\nlabel, calc, exp, error, prob"

        for lab, calc, ex, er, p in zip(slabels, svalues, sexp, serror, sprob):
            output += f"\n{lab:6s} {calc:6.2f} {ex:6.2f} {er:6.2f} {p:6.2f}"
        return output


def _load_pickle(path: str):
    """
    Loads a pickled object from relative or absolute path.
    Searches within this folder first, then within current folder, then by absolute path.

    Arguments
      path(str): path to the pickled file

    Returns
      Loaded object
    """
    _abs_path = Path(path).resolve()
    _default_path = Path(__file__).parent / path

    if _default_path.exists():
        _path = _default_path
    elif _abs_path.exists():
        _path = _abs_path
    else:
        raise FileNotFoundError("No files found at %s" % (path))
    with open(_path, "rb") as f:
        return pickle.load(f)



"""
Functions after rewrite:
NMRData structure
Should have an assignment method
"""

from .helper_functions import *
from .proton.process import proton_processing, proton_assignment
from .proton.plot import plot_proton
from .description_files import (
    process_description,
    pairwise_assignment,
    matching_assignment,
)
from .carbon.process import carbon_processing, carbon_assignment
from .carbon.plot import plot_carbon

import pickle
import logging
from pathlib import Path
from typing import List

gasConstant = 8.3145
temperature = 298.15
hartreeEnergy = 2625.499629554010
logger = logging.getLogger(__name__)


class NMRData:
    """
    must spit out the labels and shifts
    Arguments:
    nmr_source: NMR source files
    solvent: solvent used
    output_folder: path to output folder
    """

    def __init__(
        self, nmr_source: List[str], solvent: str, output_folder: Path = Path.cwd()
    ):
        self.nmr_source = [Path(i) for i in nmr_source]
        self.solvent = solvent
        self.output_folder = output_folder
        self.Atoms = []  # Element labels
        self.Cshifts = []  # Experimental C NMR shifts
        self.Clabels = []  # Experimental C NMR labels, if any
        self.Hshifts = []  # Experimental H NMR shifts
        self.Hlabels = []  # Experimental H NMR labels, if any
        self.Equivalents = (
            []
        )  # Atoms assumed to be NMR equivalent in computational data
        self.Omits = []
        self.protondata = {}
        self.carbondata = {}

        logger.info(f"Reading NMR data from {nmr_source}")

        self.search_files()
        # will guess a file. sets proton_fid and carbon_fid attributes if FID data detected

        if hasattr(self, "proton_fid"):
            self.process_proton()
        if hasattr(self, "carbon_fid"):
            self.process_carbon()

    def search_files(self):
        """Automatically searches the path for NMR data, guesses the nucleus."""
        for item in self.nmr_source:
            if item.is_dir() and (item / "fid").exists():
                logging.info("Bruker FID data found at %s" % (str(item)))
                nucleus, total_spectral_ydata, uc = read_bruker(item)
            elif item.is_file and item.suffix in (".dx", ".jdx"):
                logging.info("JCAMP-DX FID data found at %s" % (str(item)))
                nucleus, total_spectral_ydata, uc = read_jcamp(item)
            else:
                logging.info("NMR Description data found at %s" % (str(item)))
                self.process_description(item)
                return

            if nucleus == "1H":
                logger.info(f"1H NMR FID data found at: {item}")
                self.proton_fid = total_spectral_ydata, uc
            elif nucleus == "13C":
                logger.info(f"13C NMR FID data found at: {item}")
                self.carbon_fid = total_spectral_ydata, uc
        return

    def process_proton(self):
        pdir = self.output_folder / "protondata"
        gdir = self.output_folder / "graphs" / "protondata"
        if pdir.exists():
            with open(pdir, "rb") as f:
                self.protondata = pickle.load(f)
        else:
            ydata, uc = self.proton_fid
            (
                self.protondata["xdata"],
                self.protondata["ydata"],
                self.protondata["peak_regions"],
                self.protondata["grouped_peaks"],
                self.protondata["picked_peaks"],
                self.protondata["params"],
                self.protondata["sim_regions"],
            ) = proton_processing(ydata, uc, self.solvent)
            with open(pdir, "wb") as f:
                pickle.dump(self.protondata, f)

    def process_carbon(self):
        """NMR-AI not yet implemented"""
        cdir = self.output_folder / "carbondata"
        gdir = self.output_folder / "graphs" / "carbondata"
        if cdir.exists():
            with open(cdir, "rb") as f:
                self.carbondata = pickle.load(f)
        else:
            ydata, uc = self.carbon_fid

            (
                self.carbondata["xdata"],
                self.carbondata["ydata"],
                self.carbondata["exppeaks"],
                self.carbondata["simulated_ydata"],
                self.carbondata["removed"],
            ) = carbon_processing(ydata, uc, self.solvent)
            with open(cdir, "wb") as f:
                pickle.dump(self.carbondata, f)

    def process_description(self, file):

        (
            self.C_labels,
            self.C_exp,
            self.H_labels,
            self.H_exp,
            self.equivalents,
            self.omits,
        ) = process_description(file)

    def assign(self, mol):
        """
        mol: Molecule object
        Assigns data to molecule
        returns:
        - assigned experimental carbon shifts
        - scaled calculated carbon shifts
        - assigned experimental proton shifts
        - scaled calculated proton shifts
        """
        C_exp = []
        H_exp = []

        _mol = mol.rdkit_mols[0]

        C_shifts = mol.C_shifts
        C_labels = mol.C_labels
        H_shifts = mol.H_shifts
        H_labels = mol.H_labels

        if self.protondata:
            H_exp = proton_assignment(self.protondata, _mol, H_shifts, H_labels)
            plot_proton(self.protondata, self.output_folder, mol, H_exp)
        elif hasattr(self, "H_exp"):
            H_exp = pairwise_assignment(H_shifts, self.H_exp)

        if self.carbondata:
            C_exp = carbon_assignment(self.carbondata, _mol, C_shifts, C_labels)
            plot_carbon(self.carbondata, self.output_folder, mol, C_exp)
        elif hasattr(self, "C_exp"):
            C_exp = matching_assignment(C_shifts, self.C_exp, threshold=40)

        return C_exp, H_exp

    def __call__(self, mol):
        return self.assign(mol)



"""Includes baseline correction and phasing"""

import copy

import numpy as np
import nmrglue as ng
from scipy.stats import norm
from scipy.ndimage.filters import gaussian_filter1d as g1d, convolve1d as c1d
from lmfit import Minimizer, Parameters, report_fit
import statsmodels.api as sm

from ..helper_functions import normalise_intensities, lorentzian


def spectral_processing(total_spectral_ydata, uc):
    total_spectral_ydata = initial_processing(total_spectral_ydata)
    corr_distance = estimate_autocorrelation(total_spectral_ydata)
    convolved_y = gaussian_convolution(corr_distance, total_spectral_ydata)
    binary_map_regions = []

    threshold_vector = [3, 2.9, 2.8, 2.7, 2.6, 2.5, 2.2, 2, 1, 0.9, 0.8, 0.7, 0.6, 0.5]
    run = 0
    while len(binary_map_regions) < 2:
        threshold = threshold_vector[run]
        run += 1
        picked_points = iterative_point_picking(convolved_y, threshold)
        binary_map_regions, binary_map_list = binary_map(picked_points, uc, convolved_y)

    globalangles, phased_peak_regions, convolved_y_phased = estimate_phase_angles(
        convolved_y, binary_map_regions, corr_distance
    )

    real_convolved_y_phased = list(np.real(convolved_y_phased))
    picked_points_region = iterative_point_picking_region(
        binary_map_regions, real_convolved_y_phased, threshold
    )
    picked_peaks_region = peak_picking_region(
        real_convolved_y_phased, picked_points_region
    )

    p0, p1 = linear_regression(
        picked_peaks_region, globalangles, real_convolved_y_phased, binary_map_regions
    )
    total_spectral_ydata, spectral_ydata = final_phasing(convolved_y, p0, p1)
    total_spectral_ydata = total_spectral_ydata / np.max(total_spectral_ydata)

    return total_spectral_ydata, spectral_ydata, threshold, corr_distance


def initial_processing(total_spectral_ydata):
    real_part = ng.proc_bl.baseline_corrector(np.real(total_spectral_ydata), wd=2)
    im_part = ng.proc_bl.baseline_corrector(np.imag(total_spectral_ydata), wd=2)
    total_spectral_ydata = real_part + 1j * im_part
    return normalise_intensities(total_spectral_ydata)


def estimate_autocorrelation(total_spectral_ydata):

    real_part = np.real(total_spectral_ydata)
    real_part_copy = np.real(total_spectral_ydata)

    gzero = real_part * real_part
    gzero = np.sum(gzero)

    gdx = gzero

    counter = 0

    while gdx > 0.6 * gzero:
        real_part_copy = np.roll(real_part_copy, counter)
        gdx = real_part * real_part_copy
        counter += 1
        gdx = np.sum(gdx)

    corr_distance = counter

    return corr_distance


def gaussian_convolution(corr_distance, total_spectral_ydata):

    real_part = np.real(total_spectral_ydata)
    im_part = np.imag(total_spectral_ydata)

    real_convolved_y = g1d(real_part, corr_distance)
    im_convolved_y = g1d(im_part, corr_distance)

    convolved_y = np.array(real_convolved_y) + 1j * np.array(im_convolved_y)

    convolved_y = convolved_y / np.max(convolved_y)

    return convolved_y


def iterative_point_picking(convolved_y, threshold):

    real_convolved_y = np.real(convolved_y)
    copy_convolved_y = np.array(real_convolved_y)
    picked_points = []
    pickednumber = 1
    while pickednumber > 0:
        mu, std = norm.fit(copy_convolved_y)

        index = np.where(copy_convolved_y - mu > threshold * std)
        pickednumber = len(index[0])
        picked_points.extend(np.ndarray.tolist(index[0]))
        copy_convolved_y = np.delete(copy_convolved_y, index, axis=0)

    copy_convolved_y = np.array(real_convolved_y)
    pickednumber = 1
    while pickednumber > 0:
        mu, std = norm.fit(copy_convolved_y)
        index = np.where(copy_convolved_y - mu < -threshold * std)
        pickednumber = len(index[0])
        picked_points.extend(np.ndarray.tolist(index[0]))
        copy_convolved_y = np.delete(copy_convolved_y, index, axis=0)

    picked_points = sorted(picked_points)

    return picked_points


def iterative_peak_picking(
    total_spectral_ydata,
    threshold,
    corr_distance,
):

    mu, std = norm.fit(total_spectral_ydata[0:1000])

    picked_peaks = []

    # find all maxima

    maxima = []

    for point in range(1, len(total_spectral_ydata) - 1):
        if (total_spectral_ydata[point] > total_spectral_ydata[point + 1]) & (
            total_spectral_ydata[point] > total_spectral_ydata[point - 1]
        ):
            maxima.append(point)

    # start fitting process

    fit_y = np.zeros(len(total_spectral_ydata))

    while len(maxima) > 0:

        params = Parameters()

        # find peak with greatest amplitude:

        ind1 = np.argmax(total_spectral_ydata[maxima])

        peak = maxima[ind1]

        picked_peaks.append(peak)

        fit_y = minimisation(peak, fit_y, total_spectral_ydata, corr_distance)

        new_maxima = []

        for ind2 in maxima:

            if total_spectral_ydata[ind2] > threshold * std + fit_y[ind2]:

                new_maxima.append(ind2)

        maxima = copy.copy(new_maxima)

    picked_peaks = sorted(picked_peaks)

    return picked_peaks, fit_y


def binary_map(picked_points, uc, convolved_y):

    picked_points = np.array(picked_points)

    # find where peak blocks are

    binary_map_regions = [[picked_points[0]]]

    for x in range(0, len(picked_points) - 1):
        if picked_points[x + 1] != picked_points[x] + 1:
            binary_map_regions[-1].append(picked_points[x])
            binary_map_regions.append([picked_points[x + 1]])
    binary_map_regions[-1].append(picked_points[-1])

    # extend blocks by 50 Hz

    for block in binary_map_regions:
        start = uc.hz(block[0])
        end = start + 50
        end_point = uc(end, "Hz")
        block[0] = end_point

        start = uc.hz(block[1])
        end = start - 50
        end_point = uc(end, "Hz")
        block[1] = end_point

    # draw binary map

    binary_map_list = np.zeros(len(convolved_y))
    for block in binary_map_regions:
        binary_map_list[block[0] : block[1] : 1] = 1

    # stitch blocks together
    blocks = np.where(binary_map_list == 1)

    blocks = blocks[0] - 1
    binary_map_regions = [[blocks[0]]]

    for element in range(0, len(blocks) - 1):
        if blocks[element + 1] != blocks[element] + 1:
            binary_map_regions[-1].append(blocks[element])
            binary_map_regions.append([blocks[element + 1]])
    binary_map_regions[-1].append(blocks[-1])

    return binary_map_regions, binary_map_list


def minimisation(next_peak, fit_y, total_spectral_ydata, corr_distance):

    region = np.arange(
        max(0, next_peak - 100), min(next_peak + 100, len(total_spectral_ydata))
    )

    params = Parameters()

    params.add(
        "amp" + str(next_peak), value=total_spectral_ydata[next_peak], vary=False, min=0
    )
    params.add(
        "width" + str(next_peak),
        value=4 * corr_distance,
        vary=True,
        min=1 * corr_distance,
        max=8 * corr_distance,
    )
    params.add("pos" + str(next_peak), value=next_peak, vary=False)

    # print('minimising')

    out = Minimizer(
        residual,
        params,
        fcn_args=(fit_y[region], next_peak, region, total_spectral_ydata[region]),
    )

    results = out.minimize()

    # append the results params to the total params

    fit_yc = (
        lorentzian(
            np.arange(len(total_spectral_ydata)),
            results.params["width" + str(next_peak)],
            results.params["pos" + str(next_peak)],
            results.params["amp" + str(next_peak)],
        )
        + fit_y
    )

    return fit_yc


def residual(params, fit_y, next_peak, x, y_data):

    y = (
        lorentzian(
            x,
            params["width" + str(next_peak)],
            params["pos" + str(next_peak)],
            params["amp" + str(next_peak)],
        )
        + fit_y
    )

    difference = abs(y - y_data)

    return difference


def iterative_point_picking_region(
    binary_map_regions, real_convolved_y_phased, threshold
):

    copy_convolved_y = np.array(real_convolved_y_phased)
    picked_points = []
    pickednumber = 1

    while pickednumber > 0:
        mu, std = norm.fit(copy_convolved_y)
        index = np.where(
            (copy_convolved_y - mu > threshold * std)
            | (copy_convolved_y - mu < threshold * std)
        )
        picked_points.extend(np.ndarray.tolist(index[0]))
        pickednumber = len(index[0])
        copy_convolved_y = np.delete(copy_convolved_y, index, axis=0)

    picked_points = sorted(picked_points)
    picked_points_region = []

    for region in binary_map_regions:
        picked_points_region.append([])
        for point in picked_points:
            if point > region[0] and point < region[1]:
                picked_points_region[-1].append(point)

    return picked_points_region


def estimate_phase_angles(convolved_y, binary_map_regions, corr_distance):

    convolved_y_phased = np.array(convolved_y)

    def inte(binary_map_regions, peak_regions, corr_distance):
        ## for each region determine the baseline
        integrals = [0] * len(binary_map_regions)

        # first find average of surrounding points to ends of binary map regions to draw base line
        baselines_end = [[0, 0] for i in range(len(binary_map_regions))]
        baselines = [[] for i in range(0, len(binary_map_regions))]

        # find baseline endpoints
        for region in range(0, len(binary_map_regions)):
            for point in range(0, corr_distance - 1):
                baselines_end[region][0] += peak_regions[region][point]

            baselines_end[region][0] = baselines_end[region][0] / corr_distance
            for point in range(0, corr_distance):
                baselines_end[region][1] += peak_regions[region][-point]

            baselines_end[region][1] = baselines_end[region][1] / corr_distance

            ## draw baselines
            baselines[region] = np.linspace(
                baselines_end[region][0],
                baselines_end[region][1],
                len(peak_regions[region]) - 2 * corr_distance,
            )

            ## integrate each region below the baseline
            for point in range(0, len(baselines[region])):
                if (
                    peak_regions[region][point + corr_distance]
                    < baselines[region][point]
                ):
                    integrals[region] += abs(
                        peak_regions[region][point + corr_distance]
                        - baselines[region][point]
                    )

        return integrals

    coarse_angle = np.linspace(-np.pi / 2, np.pi / 2, 1000)
    integral_vector = [0] * 1000
    counter = 0
    # integration
    for angle in coarse_angle:
        copy_total_spectral_ydata = convolved_y * np.exp(-angle * 1j)
        peak_regions = [0] * len(binary_map_regions)
        for region in range(0, len(binary_map_regions)):
            peak_regions[region] = copy_total_spectral_ydata[
                binary_map_regions[region][0] : binary_map_regions[region][1] : 1
            ]

        integral_vector[counter] = inte(binary_map_regions, peak_regions, corr_distance)
        counter = counter + 1

    # find maximum integral for each region and store angles
    integral_vector = np.array(integral_vector)
    maxvector = np.amin(integral_vector, 0)

    counter = 0
    angle1 = [0] * len(binary_map_regions)

    for element in list(maxvector):
        maxangle = np.where(integral_vector == element)
        angle1[counter] = coarse_angle[maxangle[0][0]]
        counter = counter + 1

    # phase each region of the spectrum indepedently

    for region in range(0, len(peak_regions)):
        convolved_y_phased[
            binary_map_regions[region][0] : binary_map_regions[region][1] : 1
        ] = convolved_y_phased[
            binary_map_regions[region][0] : binary_map_regions[region][1] : 1
        ] * np.exp(
            -angle1[region] * 1j
        )

    globalangles = [angle1[i] for i in range(0, len(binary_map_regions))]

    ## phase each peak region separately

    phased_peak_regions = []
    copy_total_spectral_ydata = convolved_y
    peak_regions = [0] * len(binary_map_regions)
    for region in range(0, len(binary_map_regions)):
        peak_regions[region] = copy_total_spectral_ydata[
            binary_map_regions[region][0] : binary_map_regions[region][1] : 1
        ]
    counter = 0

    for region in peak_regions:
        phased_peak_regions.append(region * np.exp(-globalangles[counter] * 1j))
        counter += 1

    return globalangles, phased_peak_regions, convolved_y_phased


def peak_picking_region(real_convolved_y_phased, picked_points_region):
    picked_peaks_region = []

    for region in range(0, len(picked_points_region)):
        picked_peaks_region.append([])

        for index in picked_points_region[region]:
            peak = real_convolved_y_phased[index]
            if (
                peak > real_convolved_y_phased[index + 1]
                and peak > real_convolved_y_phased[index - 1]
            ):
                picked_peaks_region[-1].append(index)
            elif (
                peak < real_convolved_y_phased[index + 1]
                and peak < real_convolved_y_phased[index - 1]
            ):
                picked_peaks_region[-1].append(index)
    return picked_peaks_region


def linear_regression(
    picked_peaks_region, globalangles, real_convolved_y_phased, binary_map_regions
):

    # region weighting vector

    region_weighting_matrix = [1] * len(picked_peaks_region)

    for index, region in enumerate(picked_peaks_region):
        region_weighting_matrix[index] = max(
            [abs(real_convolved_y_phased[peak]) for peak in region]
        )
    max_weight = max(region_weighting_matrix)

    region_weighting_matrix = [i / max_weight for i in region_weighting_matrix]

    # define centres of regions
    region_centres = []

    for region in binary_map_regions:
        region_centres.append(
            (1 - (region[0] + region[1]) / (2 * len(real_convolved_y_phased)))
        )

    #### regression and outlier analysis

    number_of_outliers = 1
    while number_of_outliers > 0:
        region_centres_regression = sm.add_constant(region_centres)
        wls_model = sm.WLS(
            globalangles, region_centres_regression, weights=region_weighting_matrix
        )
        results = wls_model.fit()
        params = results.params
        predictions = [params[1] * i + params[0] for i in region_centres]
        # remove maximum outlier more than 0.6 rad from estimate
        differences = [
            abs(predictions[angle] - globalangles[angle])
            for angle in range(0, len(globalangles))
        ]

        maxdifference = max(differences)
        if maxdifference > 0.6:
            index = differences.index(maxdifference)
            globalangles.pop(index)
            region_centres.pop(index)
            region_weighting_matrix.pop(index)
        else:
            number_of_outliers = 0

    p0 = params[0]
    p1 = params[1]

    return p0, p1


def final_phasing(total_spectral_ydata, p0, p1):

    # total_spectral_ydata = ng.proc_base.ps(total_spectral_ydata, p0=p0, p1=p1)

    relativeposition = np.linspace(1, 0, len(total_spectral_ydata))

    angle = p0 + p1 * relativeposition

    total_spectral_ydata = total_spectral_ydata * np.exp(-1j * angle)

    total_spectral_ydata = ng.proc_bl.baseline_corrector(total_spectral_ydata, wd=2)

    spectral_ydata = ng.proc_base.di(total_spectral_ydata)  # discard the imaginaries
    spectral_ydata = np.ndarray.tolist(spectral_ydata)
    total_spectral_ydata = np.real(total_spectral_ydata)

    return total_spectral_ydata, spectral_ydata



import itertools

import numpy as np
from lmfit import Parameters

from dp5.nmr_processing.proton.pearson7 import p7sim, p7plot
from dp5.nmr_processing.helper_functions import lorentzian


def simulate_regions(params, peak_regions, grouped_peaks, y_data, xppm):
    sim_regions = np.empty(len(peak_regions), dtype=object)
    sim_y = np.zeros(len(y_data))

    for ind, region in enumerate(peak_regions):
        sim_y[region] = p7plot(params, region, grouped_peaks[ind], ind, xppm)

        y = p7sim(params, region, grouped_peaks[ind], ind)

        sim_regions[ind] = y

    return sim_regions, sim_y


def new_first_order_peak(start_ppm, J_vals, x_data, corr_distance, uc, spin):
    # new first order peak generator using the method presented in Hoye paper

    start = uc(str(start_ppm) + "ppm")

    start_Hz = uc.hz(start)

    J_vals = np.array(J_vals)

    peaks = np.zeros((2 * spin + 1) ** len(J_vals))

    if spin == 0.5:
        l = [1, -1]

    if spin == 1:
        l = [1, 0, -1]

    # signvector generator

    signvectors = itertools.product(l, repeat=len(J_vals))

    for ind, sv in enumerate(signvectors):
        shift = J_vals * sv

        shift = start_Hz + np.sum(shift)
        peaks[ind] = shift

    peaks = np.sort(peaks)

    peak_vector = np.array(sorted(list(set(peaks)), reverse=True))

    amp_vector = np.zeros(len(peak_vector))

    for peak in peaks:
        index = np.where(peak_vector == peak)
        amp_vector[index] += 1

    pv = []

    for index, peak in enumerate(peak_vector):
        pv.append(uc(peak, "Hz"))

    peak_vector = pv

    split_params = Parameters()

    for index, peak in enumerate(peak_vector):
        split_params.add("amp" + str(peak), value=amp_vector[index])
        split_params.add("pos" + str(peak), value=peak)
        split_params.add("width" + str(peak), value=2 * corr_distance)

    y = lorenz_curves(split_params, x_data, peak_vector)
    y = y / np.max(y)

    # where = np.where(y > 0.001)
    # y = y[where]

    return peak_vector, amp_vector, y


def lorenz_curves(params, x, picked_points):
    y = np.zeros(len(x))
    for peak in picked_points:
        y += lorentzian(
            x,
            params["width" + str(peak)],
            params["pos" + str(peak)],
            params["amp" + str(peak)],
        )
    return y



import logging, sys

import numpy as np
from tqdm import tqdm
from scipy.linalg import eigh

from rdkit import Chem
from rdkit.Chem import MolFromSmiles, MolToSmiles, AddHs

from nfp.preprocessing import features
from nfp.preprocessing.features import Tokenizer
import time


class SmilesPreprocessor(object):
    """ Given a list of SMILES strings, encode these molecules as atom and
    connectivity feature matricies.

    Example:
    >>> preprocessor = SmilesPreprocessor(explicit_hs=False)
    >>> inputs = preprocessor.fit(data.smiles)
    """

    def __init__(self, explicit_hs=True, atom_features=None, bond_features=None):
        """

        explicit_hs : bool
            whether to tell RDkit to add H's to a molecule.
        atom_features : function
            A function applied to an rdkit.Atom that returns some
            representation (i.e., string, integer) for the Tokenizer class.
        bond_features : function
            A function applied to an rdkit Bond to return some description.

        """

        self.atom_tokenizer = Tokenizer()
        self.bond_tokenizer = Tokenizer()
        self.explicit_hs = explicit_hs

        if atom_features is None:
            atom_features = features.atom_features_v1

        if bond_features is None:
            bond_features = features.bond_features_v1

        self.atom_features = atom_features
        self.bond_features = bond_features


    def fit(self, smiles_iterator):
        """ Fit an iterator of SMILES strings, creating new atom and bond
        tokens for unseen molecules. Returns a dictionary with 'atom' and
        'connectivity' entries """
        return list(self.preprocess(smiles_iterator, train=True))


    def predict(self, smiles_iterator):
        """ Uses previously determined atom and bond tokens to convert a SMILES
        iterator into 'atom' and 'connectivity' matrices. Ensures that atom and
        bond classes commute with previously determined results. """
        return list(self.preprocess(smiles_iterator, train=False))


    def preprocess(self, smiles_iterator, train=True):

        self.atom_tokenizer.train = train
        self.bond_tokenizer.train = train

        for smiles in tqdm(smiles_iterator):
            yield self.construct_feature_matrices(smiles)


    @property
    def atom_classes(self):
        """ The number of atom types found (includes the 0 null-atom type) """
        return self.atom_tokenizer.num_classes + 1


    @property
    def bond_classes(self):
        """ The number of bond types found (includes the 0 null-bond type) """
        return self.bond_tokenizer.num_classes + 1


    def construct_feature_matrices(self, smiles):
        """ construct a molecule from the given smiles string and return atom
        and bond classes.

        Returns
        dict with entries
        'n_atom' : number of atoms in the molecule
        'n_bond' : number of bonds in the molecule 
        'atom' : (n_atom,) length list of atom classes
        'bond' : (n_bond,) list of bond classes
        'connectivity' : (n_bond, 2) array of source atom, target atom pairs.

        """

        mol = MolFromSmiles(smiles)
        if self.explicit_hs:
            mol = AddHs(mol)

        n_atom = len(mol.GetAtoms())
        n_bond = 2 * len(mol.GetBonds())

        # If its an isolated atom, add a self-link
        if n_bond == 0:
            n_bond = 1
        
        atom_feature_matrix = np.zeros(n_atom, dtype='int')
        bond_feature_matrix = np.zeros(n_bond, dtype='int')
        connectivity = np.zeros((n_bond, 2), dtype='int')

        bond_index = 0

        atom_seq = mol.GetAtoms()
        atoms = [atom_seq[i] for i in range(n_atom)]

        for n, atom in enumerate(atoms):

            # Atom Classes
            atom_feature_matrix[n] = self.atom_tokenizer(
                self.atom_features(atom))

            start_index = atom.GetIdx()

            for bond in atom.GetBonds():
                # Is the bond pointing at the target atom
                rev = bond.GetBeginAtomIdx() != start_index

                # Bond Classes
                bond_feature_matrix[n] = self.bond_tokenizer(
                    self.bond_features(bond, flipped=rev))

                # Connectivity
                if not rev:  # Original direction
                    connectivity[bond_index, 0] = bond.GetBeginAtomIdx()
                    connectivity[bond_index, 1] = bond.GetEndAtomIdx()

                else:  # Reversed
                    connectivity[bond_index, 0] = bond.GetEndAtomIdx()
                    connectivity[bond_index, 1] = bond.GetBeginAtomIdx()

                bond_index += 1


        return {
            'n_atom': n_atom,
            'n_bond': n_bond,
            'atom': atom_feature_matrix,
            'bond': bond_feature_matrix,
            'connectivity': connectivity,
        }
    

class ConnectivityAPreprocessor(object):
    """ Given a list of SMILES strings, encode these molecules as atom and
    connectivity feature matricies.

    Example:
    >>> preprocessor = SmilesPreprocessor(explicit_hs=False)
    >>> inputs = preprocessor.fit(data.smiles)
    """

    def __init__(self, explicit_hs=True, atom_features=None, bond_features=None):
        """

        explicit_hs : bool
            whether to tell RDkit to add H's to a molecule.
        atom_features : function
            A function applied to an rdkit.Atom that returns some
            representation (i.e., string, integer) for the Tokenizer class.
        bond_features : function
            A function applied to an rdkit Bond to return some description.

        """

        self.atom_tokenizer = Tokenizer()
        self.bond_tokenizer = Tokenizer()
        self.explicit_hs = explicit_hs

        if atom_features is None:
            atom_features = features.atom_features_v1

        if bond_features is None:
            bond_features = features.bond_features_v1

        self.atom_features = atom_features
        self.bond_features = bond_features


    def fit(self, smiles_iterator):
        """ Fit an iterator of SMILES strings, creating new atom and bond
        tokens for unseen molecules. Returns a dictionary with 'atom' and
        'connectivity' entries """
        return list(self.preprocess(smiles_iterator, train=True))


    def predict(self, smiles_iterator):
        """ Uses previously determined atom and bond tokens to convert a SMILES
        iterator into 'atom' and 'connectivity' matrices. Ensures that atom and
        bond classes commute with previously determined results. """
        return list(self.preprocess(smiles_iterator, train=False))


    def preprocess(self, smiles_iterator, train=True):

        self.atom_tokenizer.train = train
        self.bond_tokenizer.train = train

        for smiles in tqdm(smiles_iterator):
            yield self.construct_feature_matrices(smiles)


    @property
    def atom_classes(self):
        """ The number of atom types found (includes the 0 null-atom type) """
        return self.atom_tokenizer.num_classes + 1


    @property
    def bond_classes(self):
        """ The number of bond types found (includes the 0 null-bond type) """
        return self.bond_tokenizer.num_classes + 1


    def construct_feature_matrices(self, smiles):
        """ construct a molecule from the given smiles string and return atom
        and bond classes.

        Returns
        dict with entries
        'n_atom' : number of atoms in the molecule
        'n_bond' : number of bonds in the molecule 
        'atom' : (n_atom,) length list of atom classes
        'bond' : (n_bond,) list of bond classes
        'connectivity' : (n_bond, 2) array of source atom, target atom pairs.

        """

        mol = MolFromSmiles(smiles)
        if self.explicit_hs:
            mol = AddHs(mol)

        n_atom = len(mol.GetAtoms())
        n_bond = 2 * len(mol.GetBonds())

        # If its an isolated atom, add a self-link
        if n_bond == 0:
            n_bond = 1
        
        atom_feature_matrix = np.zeros(n_atom, dtype='int')
        bond_feature_matrix = np.zeros(n_bond, dtype='int')
        connectivity = np.zeros((n_bond, 2), dtype='int')

        bond_index = 0

        atom_seq = mol.GetAtoms()
        atoms = [atom_seq[i] for i in range(n_atom)]

        for n, atom in enumerate(atoms):

            # Atom Classes
            atom_feature_matrix[n] = self.atom_tokenizer(
                self.atom_features(atom))

            start_index = atom.GetIdx()

            for bond in atom.GetBonds():
                # Is the bond pointing at the target atom
                rev = bond.GetBeginAtomIdx() != start_index

                # Bond Classes
                bond_feature_matrix[n] = self.bond_tokenizer(
                    self.bond_features(bond, flipped=rev))

                # Connectivity
                if not rev:  # Original direction
                    connectivity[bond_index, 0] = bond.GetBeginAtomIdx()
                    connectivity[bond_index, 1] = bond.GetEndAtomIdx()

                else:  # Reversed
                    connectivity[bond_index, 0] = bond.GetEndAtomIdx()
                    connectivity[bond_index, 1] = bond.GetBeginAtomIdx()

                bond_index += 1


        return {
            'n_atom': n_atom,
            'n_bond': n_bond,
            'atom': atom_feature_matrix,
            'bond': bond_feature_matrix,
            'connectivity': connectivity,
        }


class MolPreprocessor(SmilesPreprocessor):
    """ I should refactor this into a base class and separate
    SmilesPreprocessor classes. But the idea is that we only need to redefine
    the `construct_feature_matrices` method to have a working preprocessor that
    handles 3D structures. 

    We'll pass an iterator of mol objects instead of SMILES strings this time,
    though.
    
    """

    def __init__(self, n_neighbors, cutoff, **kwargs):
        """ A preprocessor class that also returns distances between
        neighboring atoms. Adds edges for non-bonded atoms to include a maximum
        of n_neighbors around each atom """

        self.n_neighbors = n_neighbors
        self.cutoff = cutoff
        super(MolPreprocessor, self).__init__(**kwargs)


    def construct_feature_matrices(self, mol):
        """ Given an rdkit mol, return atom feature matrices, bond feature
        matrices, and connectivity matrices.

        Returns
        dict with entries
        'n_atom' : number of atoms in the molecule
        'n_bond' : number of edges (likely n_atom * n_neighbors)
        'atom' : (n_atom,) length list of atom classes
        'bond' : (n_bond,) list of bond classes. 0 for no bond
        'distance' : (n_bond,) list of bond distances
        'connectivity' : (n_bond, 2) array of source atom, target atom pairs.
            
        """

        n_atom = len(mol.GetAtoms())

        # n_bond is actually the number of atom-atom pairs, so this is defined
        # by the number of neighbors for each atom.
        #if there is cutoff, 
        distance_matrix = Chem.Get3DDistanceMatrix(mol)

        if self.n_neighbors <= (n_atom - 1):
            n_bond = self.n_neighbors * n_atom
        else:
            # If there are fewer atoms than n_neighbors, all atoms will be
            # connected
            n_bond = distance_matrix[(distance_matrix < self.cutoff) & (distance_matrix != 0)].size

        if n_bond == 0: n_bond = 1

        # Initialize the matrices to be filled in during the following loop.
        atom_feature_matrix = np.zeros(n_atom, dtype='int')
        bond_feature_matrix = np.zeros(n_bond, dtype='int')
        bond_distance_matrix = np.zeros(n_bond, dtype=np.float32)
        connectivity = np.zeros((n_bond, 2), dtype='int')

        # Hopefully we've filtered out all problem mols by now.
        if mol is None:
            raise RuntimeError("Issue in loading mol")
        
        # Get a list of the atoms in the molecule.
        atom_seq = mol.GetAtoms()
        atoms = [atom_seq[i] for i in range(n_atom)]

        # Here we loop over each atom, and the inner loop iterates over each
        # neighbor of the current atom.
        bond_index = 0  # keep track of our current bond.
        for n, atom in enumerate(atoms):
            
            # update atom feature matrix
            atom_feature_matrix[n] = self.atom_tokenizer(
                self.atom_features(atom))
            
            # if n_neighbors is greater than total atoms, then each atom is a
            # neighbor.
            if (self.n_neighbors + 1) > len(mol.GetAtoms()):
                neighbor_end_index = len(mol.GetAtoms())
            else:
                neighbor_end_index = (self.n_neighbors + 1)

            distance_atom = distance_matrix[n, :]
            cutoff_end_index = distance_atom[distance_atom < self.cutoff].size

            end_index = min(neighbor_end_index, cutoff_end_index)

            # Loop over each of the nearest neighbors

            neighbor_inds = distance_matrix[n, :].argsort()[1:end_index]
            if len(neighbor_inds)==0: neighbor_inds = [n]
            for neighbor in neighbor_inds:
                
                # update bond feature matrix
                bond = mol.GetBondBetweenAtoms(n, int(neighbor))
                if bond is None:
                    bond_feature_matrix[bond_index] = 0
                else:
                    rev = False if bond.GetBeginAtomIdx() == n else True
                    bond_feature_matrix[bond_index] = self.bond_tokenizer(
                        self.bond_features(bond, flipped=rev))

                distance = distance_matrix[n, neighbor]
                bond_distance_matrix[bond_index] = distance
                
                # update connectivity matrix
                connectivity[bond_index, 0] = n
                connectivity[bond_index, 1] = neighbor
                
                bond_index += 1

        return {
            'n_atom': n_atom,
            'n_bond': n_bond,
            'atom': atom_feature_matrix,
            'bond': bond_feature_matrix,
            'distance': bond_distance_matrix,
            'connectivity': connectivity,
        }


class MolBPreprocessor(MolPreprocessor):
    """
    This is a subclass of Molpreprocessor that preprocessor molecule with
    bond property target
    """
    def __init__(self, **kwargs):
        """
        A preprocessor class that also returns bond_target_matrix, besides the bond matrix
        returned by MolPreprocessor. The bond_target_matrix is then used as ref to reduce molecule
        to bond property
        """
        super(MolBPreprocessor, self).__init__(**kwargs)

    def construct_feature_matrices(self, entry):
        """
        Given an entry contining rdkit molecule, bond_index and for the target property, 
        return atom 
        feature matrices, bond feature matrices, distance matrices, connectivity matrices and bond
        ref matrices.

        returns
        dict with entries
        see MolPreproccessor
        'bond_index' : ref array to the bond index
        """
        mol, bond_index_array = entry
        
        n_atom = len(mol.GetAtoms())
        n_pro = len(bond_index_array)

        # n_bond is actually the number of atom-atom pairs, so this is defined
        # by the number of neighbors for each atom.
        #if there is cutoff, 
        distance_matrix = Chem.Get3DDistanceMatrix(mol)

        if self.n_neighbors <= (n_atom - 1):
            n_bond = self.n_neighbors * n_atom
        else:
            # If there are fewer atoms than n_neighbors, all atoms will be
            # connected
            n_bond = distance_matrix[(distance_matrix < self.cutoff) & (distance_matrix != 0)].size

        if n_bond == 0: n_bond = 1

        # Initialize the matrices to be filled in during the following loop.
        atom_feature_matrix = np.zeros(n_atom, dtype='int')
        bond_feature_matrix = np.zeros(n_bond, dtype='int')
        bond_distance_matrix = np.zeros(n_bond, dtype=np.float32)
        bond_index_matrix = np.full(n_bond, -1, dtype='int')
        connectivity = np.zeros((n_bond, 2), dtype='int')

        # Hopefully we've filtered out all problem mols by now.
        if mol is None:
            raise RuntimeError("Issue in loading mol")
        
        # Get a list of the atoms in the molecule.
        atom_seq = mol.GetAtoms()
        atoms = [atom_seq[i] for i in range(n_atom)]

        # Here we loop over each atom, and the inner loop iterates over each
        # neighbor of the current atom.
        bond_index = 0  # keep track of our current bond.
        for n, atom in enumerate(atoms):
            # update atom feature matrix
            atom_feature_matrix[n] = self.atom_tokenizer(
                self.atom_features(atom))
            
            # if n_neighbors is greater than total atoms, then each atom is a
            # neighbor.
            if (self.n_neighbors + 1) > len(mol.GetAtoms()):
                neighbor_end_index = len(mol.GetAtoms())
            else:
                neighbor_end_index = (self.n_neighbors + 1)

            distance_atom = distance_matrix[n, :]
            cutoff_end_index = distance_atom[distance_atom < self.cutoff].size

            end_index = min(neighbor_end_index, cutoff_end_index)

            # Loop over each of the nearest neighbors

            neighbor_inds = distance_matrix[n, :].argsort()[1:end_index]
            if len(neighbor_inds)==0: neighbor_inds = [n]
            for neighbor in neighbor_inds:
                
                # update bond feature matrix
                bond = mol.GetBondBetweenAtoms(n, int(neighbor))
                if bond is None:
                    bond_feature_matrix[bond_index] = 0
                else:
                    rev = False if bond.GetBeginAtomIdx() == n else True
                    bond_feature_matrix[bond_index] = self.bond_tokenizer(
                        self.bond_features(bond, flipped=rev))
                    try:
                        bond_index_matrix[bond_index] = bond_index_array.tolist().index(bond.GetIdx())
                    except:
                        pass

                distance = distance_matrix[n, neighbor]
                bond_distance_matrix[bond_index] = distance
                 
                # update connectivity matrix
                connectivity[bond_index, 0] = n
                connectivity[bond_index, 1] = neighbor
                
                bond_index += 1
        return {
            'n_atom': n_atom,
            'n_bond': n_bond,
            'n_pro': n_pro,
            'atom': atom_feature_matrix,
            'bond': bond_feature_matrix,
            'distance': bond_distance_matrix,
            'connectivity': connectivity,
            'bond_index': bond_index_matrix,
        }

class MolAPreprocessor(MolPreprocessor):
    """
    This is a subclass of Molpreprocessor that preprocessor molecule with
    bond property target
    """
    def __init__(self, **kwargs):
        """
        A preprocessor class that also returns bond_target_matrix, besides the bond matrix
        returned by MolPreprocessor. The bond_target_matrix is then used as ref to reduce molecule
        to bond property
        """
        super(MolAPreprocessor, self).__init__(**kwargs)

    def construct_feature_matrices(self, entry):
        """
        Given an entry contining rdkit molecule, bond_index and for the target property, 
        return atom 
        feature matrices, bond feature matrices, distance matrices, connectivity matrices and bond
        ref matrices.

        returns
        dict with entries
        see MolPreproccessor
        'bond_index' : ref array to the bond index
        """
        mol, atom_index_array = entry
        
        n_atom = len(mol.GetAtoms())
        n_pro = len(atom_index_array)

        # n_bond is actually the number of atom-atom pairs, so this is defined
        # by the number of neighbors for each atom.
        #if there is cutoff, 
        distance_matrix = Chem.Get3DDistanceMatrix(mol)

        if self.n_neighbors <= (n_atom - 1):
            n_bond = self.n_neighbors * n_atom
        else:
            # If there are fewer atoms than n_neighbors, all atoms will be
            # connected
            n_bond = distance_matrix[(distance_matrix < self.cutoff) & (distance_matrix != 0)].size

        if n_bond == 0: n_bond = 1

        # Initialize the matrices to be filled in during the following loop.
        atom_feature_matrix = np.zeros(n_atom, dtype='int')
        bond_feature_matrix = np.zeros(n_bond, dtype='int')
        bond_distance_matrix = np.zeros(n_bond, dtype=np.float32)
        atom_index_matrix = np.full(n_atom, -1, dtype='int')
        connectivity = np.zeros((n_bond, 2), dtype='int')

        # Hopefully we've filtered out all problem mols by now.
        if mol is None:
            raise RuntimeError("Issue in loading mol")
        
        # Get a list of the atoms in the molecule.
        atom_seq = mol.GetAtoms()
        atoms = [atom_seq[i] for i in range(n_atom)]

        # Here we loop over each atom, and the inner loop iterates over each
        # neighbor of the current atom.
        bond_index = 0  # keep track of our current bond.
        for n, atom in enumerate(atoms):
            # update atom feature matrix
            atom_feature_matrix[n] = self.atom_tokenizer(
                self.atom_features(atom))
            try:
                atom_index_matrix[n] = atom_index_array.tolist().index(atom.GetIdx())
            except:
                pass 
            # if n_neighbors is greater than total atoms, then each atom is a
            # neighbor.
            if (self.n_neighbors + 1) > len(mol.GetAtoms()):
                neighbor_end_index = len(mol.GetAtoms())
            else:
                neighbor_end_index = (self.n_neighbors + 1)

            distance_atom = distance_matrix[n, :]
            cutoff_end_index = distance_atom[distance_atom < self.cutoff].size

            end_index = min(neighbor_end_index, cutoff_end_index)

            # Loop over each of the nearest neighbors

            neighbor_inds = distance_matrix[n, :].argsort()[1:end_index]
            if len(neighbor_inds)==0: neighbor_inds = [n]
            for neighbor in neighbor_inds:
                
                # update bond feature matrix
                bond = mol.GetBondBetweenAtoms(n, int(neighbor))
                if bond is None:
                    bond_feature_matrix[bond_index] = 0
                else:
                    rev = False if bond.GetBeginAtomIdx() == n else True
                    bond_feature_matrix[bond_index] = self.bond_tokenizer(
                        self.bond_features(bond, flipped=rev))

                distance = distance_matrix[n, neighbor]
                bond_distance_matrix[bond_index] = distance
                 
                # update connectivity matrix
                connectivity[bond_index, 0] = n
                connectivity[bond_index, 1] = neighbor
                
                bond_index += 1
        return {
            'n_atom': n_atom,
            'n_bond': n_bond,
            'n_pro': n_pro,
            'atom': atom_feature_matrix,
            'bond': bond_feature_matrix,
            'distance': bond_distance_matrix,
            'connectivity': connectivity,
            'atom_index': atom_index_matrix,
        }


# TODO: rewrite this                                
# class LaplacianSmilesPreprocessor(SmilesPreprocessor):
#     """ Extends the SmilesPreprocessor class to also return eigenvalues and
#     eigenvectors of the graph laplacian matrix.
#
#     Example:
#     >>> preprocessor = SmilesPreprocessor(
#     >>>     max_atoms=55, max_bonds=62, max_degree=4, explicit_hs=False)
#     >>> atom, connectivity, eigenvalues, eigenvectors = preprocessor.fit(
#             data.smiles)
#     """
#
#     def preprocess(self, smiles_iterator, train=True):
#
#         self.atom_tokenizer.train = train
#         self.bond_tokenizer.train = train
#
#         for smiles in tqdm(smiles_iterator):
#             G = self._mol_to_nx(smiles)
#             A = self._get_atom_feature_matrix(G)
#             C = self._get_connectivity_matrix(G)
#             W, V = self._get_laplacian_spectral_decomp(G)
#             yield A, C, W, V
#
#
#     def _get_laplacian_spectral_decomp(self, G):
#         """ Return the eigenvalues and eigenvectors of the graph G, padded to
#         `self.max_atoms`.
#         """
#
#         w0 = np.zeros((self.max_atoms, 1))
#         v0 = np.zeros((self.max_atoms, self.max_atoms))
#
#         w, v = eigh(nx.laplacian_matrix(G).todense())
#
#         num_atoms = len(v)
#
#         w0[:num_atoms, 0] = w
#         v0[:num_atoms, :num_atoms] = v
#
#         return w0, v0
#
#
#     def fit(self, smiles_iterator):
#         results = self._fit(smiles_iterator)
#         return {'atom': results[0], 
#                 'connectivity': results[1],
#                 'w': results[2],
#                 'v': results[3]}
#
#
#     def predict(self, smiles_iterator):
#         results = self._predict(smiles_iterator)
#         return {'atom': results[0], 
#                 'connectivity': results[1],
#                 'w': results[2],
#                 'v': results[3]}


def get_max_atom_bond_size(smiles_iterator, explicit_hs=True):
    """ Convienence function to get max_atoms, max_bonds for a set of input
    SMILES """

    max_atoms = 0
    max_bonds = 0
    for smiles in tqdm(smiles_iterator):
        mol = MolFromSmiles(smiles)
        if explicit_hs:
            mol = AddHs(mol)
        max_atoms = max([max_atoms, len(mol.GetAtoms())])
        max_bonds = max([max_bonds, len(mol.GetBonds())])

    return dict(max_atoms=max_atoms, max_bonds=max_bonds*2)


def canonicalize_smiles(smiles, isomeric=True, sanitize=True):
    try:
        mol = MolFromSmiles(smiles, sanitize=sanitize)
        return MolToSmiles(mol, isomericSmiles=isomeric)
    except Exception:
        pass



import sys
import os

sys.path.insert(0, os.path.abspath(os.path.join("..", "..")))
# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "DP5"
copyright = "2024, Kristaps Ermanis, Alexander Howarth, Ruslan Kotlyarov, Benji Rowlands, Jonathan Goodman"
author = "Kristaps Ermanis, Alexander Howarth, Ruslan Kotlyarov, Benji Rowlands, Jonathan Goodman"
release = "0.1"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.napoleon",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.githubpages",
]

templates_path = ["_templates"]
exclude_patterns = []
autodoc_mock_imports = [
    "numpy",
    "sklearn",
    "pandas",
    "tensorflow",
    "pathos",
    "scipy",
    "nmrglue",
    "matplotlib",
    "tomli",
    "lmfit",
    "openbabel",
    "statsmodels",
    "rdkit",
    "tqdm",
    "keras",
]
# autodoc_typehints = "description"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



import logging
import time

import numpy as np

from dp5.nmr_processing.proton import (
    find_integrals,
    gradient_peak_picking,
    spectral_processing,
    solvent_removal,
    multiproc_BIC_minimisation,
    simulate_regions,
    weighted_region_centres,
    iterative_assignment,
)


logger = logging.getLogger(__name__)


def proton_processing(total_spectral_ydata, uc, solvent):
    """
    Handles the initial processing, peak picking, solvent removal and integration.
    Should not handle impurities until assignment
    arguments:
        - total_spectral_ydata: complex array for FID data
        – uc: nmrglue unit conversion object
        - solvent: required for solvent removal

    returns:
        - spectral_xdata_ppm
        - total_spectral_ydata
        - peak_regions
        - grouped_peaks
        - solvent_region_ind
        - picked_peaks
        - total_params
        - sim_regions
    """
    (
        total_spectral_ydata,
        corr_distance,
        noise_std,
        peak_regions,
    ) = spectral_processing(total_spectral_ydata)

    gradient_peaks, gradient_regions, gradient_groups, std = gradient_peak_picking(
        total_spectral_ydata, corr_distance, uc, noise_std, peak_regions
    )

    start = time.time()

    picked_peaks, grouped_peaks, peak_regions, sim_y, total_params = (
        multiproc_BIC_minimisation(
            gradient_regions,
            gradient_groups,
            total_spectral_ydata,
            corr_distance,
            uc,
            noise_std,
        )
    )

    end = time.time()

    logger.info(f"minimisation time = {(end - start) / 60:.2f} mins")

    spectral_xdata_ppm = uc.ppm_scale()

    (
        peak_regions,
        picked_peaks,
        grouped_peaks,
        spectral_xdata_ppm,
        solvent_region_ind,
    ) = solvent_removal(
        solvent,
        total_spectral_ydata,
        spectral_xdata_ppm,
        picked_peaks,
        peak_regions,
        grouped_peaks,
        total_params,
        uc,
    )

    sim_regions, full_sim_data = simulate_regions(
        total_params,
        peak_regions,
        grouped_peaks,
        total_spectral_ydata,
        spectral_xdata_ppm,
    )

    return (
        spectral_xdata_ppm,
        total_spectral_ydata,
        peak_regions,
        grouped_peaks,
        picked_peaks,
        total_params,
        sim_regions,
    )


def integrate_and_remove_impurities(
    mol,
    spectral_xdata_ppm,
    total_spectral_ydata,
    peak_regions,
    grouped_peaks,
    picked_peaks,
    total_params,
    sim_regions,
):
    """Takes the molecular structure, identifies relevant peaks"""
    (
        peak_regions,
        grouped_peaks,
        sim_regions,
        integral_sum,
        cummulative_vectors,
        integrals,
        number_of_protons_structure,
        optimum_proton_number,
        total_integral,
    ) = find_integrals(
        mol,
        peak_regions,
        grouped_peaks,
        sim_regions,
        picked_peaks,
        total_params,
        total_spectral_ydata,
    )

    # find region centres

    centres = weighted_region_centres(peak_regions, total_spectral_ydata)

    ################

    exp_peaks = []

    integrals = np.array([int(i) for i in integrals])

    for ind, peak in enumerate(centres):
        exp_peaks += [peak] * integrals[ind]

    integrals = integrals[integrals > 0.5]

    exp_peaks = spectral_xdata_ppm[exp_peaks]

    exp_peaks = np.array([round(i, 4) for i in exp_peaks])
    # it is best we standardise everyting as array of arrays for compatibility with subsequent steps
    return (
        exp_peaks,
        spectral_xdata_ppm,
        total_spectral_ydata,
        integrals,
        peak_regions,
        centres,
        cummulative_vectors,
        integral_sum,
        picked_peaks,
        total_params,
        sim_regions,
    )


def proton_assignment(nmr_data, molecule, shifts, labels) -> list:
    """
    Handles final round of impurity removal
    arguments:
    - NMRData object
    - molecule(rdkit.Mol): rdkit Mol object to track the atom numbering
    - shifts(list): list of predicted shifts
    - labels(list) list of proton labels
    returns:
    list of experimetal peak values, same order as in shifts
    """
    (
        exp_peaks,
        spectral_xdata_ppm,
        total_spectral_ydata,
        integrals,
        peak_regions,
        centres,
        cummulative_vectors,
        integral_sum,
        picked_peaks,
        total_params,
        sim_regions,
    ) = integrate_and_remove_impurities(
        mol=molecule,
        spectral_xdata_ppm=nmr_data["xdata"],
        total_spectral_ydata=nmr_data["ydata"],
        peak_regions=nmr_data["peak_regions"],
        grouped_peaks=nmr_data["grouped_peaks"],
        picked_peaks=nmr_data["picked_peaks"],
        total_params=nmr_data["params"],
        sim_regions=nmr_data["sim_regions"],
    )
    assigned_shifts, assigned_peaks, assigned_labels, scaled_shifts = (
        iterative_assignment(
            mol=molecule,
            exp_peaks=exp_peaks,
            calculated_shifts=shifts,
            H_labels=labels,
            rounded_integrals=integrals,
        )
    )
    H_exp = [None] * len(shifts)
    for label, peak in zip(assigned_labels, assigned_peaks):

        w = labels.tolist().index(label)

        H_exp[w] = peak
    return np.array(H_exp)



import numpy as np
from scipy.stats import norm
from matplotlib import pyplot as plt
from scipy.optimize import linear_sum_assignment as optimise
from scipy.stats import linregress, gaussian_kde
import copy

from rdkit import Chem


def iterative_assignment(
    picked_peaks, spectral_xdata_ppm, total_spectral_ydata, calculated_shifts, C_labels
):
    calculated_shifts = np.array(calculated_shifts)

    original_C_labels = np.array(C_labels)

    s = np.argsort(np.array(calculated_shifts))

    calculated_shifts = calculated_shifts[s]

    scaled_shifts = copy.copy(calculated_shifts)

    C_labels = original_C_labels[s]

    exp_peaks = spectral_xdata_ppm[picked_peaks]

    new_assigned_peaks = []

    new_assigned_shifts = []

    for lnum in range(0, 2):

        if lnum == 0:

            scaled_shifts = external_scale_carbon_shifts(calculated_shifts)

            scaled_mu = 0

            scaled_std = 2.486068603518297

            copy_calc_shifts = copy.copy(calculated_shifts)

        elif lnum == 1:

            old_assigned_shifts = copy.copy(new_assigned_shifts)

            old_assigned_peaks = copy.copy(new_assigned_peaks)

            scaled_shifts, slope, intercept = internal_scale_carbon_shifts(
                old_assigned_shifts, old_assigned_peaks, calculated_shifts
            )
            scaled_mu = 0

            scaled_std = 10

            copy_calc_shifts = copy.copy(calculated_shifts)

        ####calculate difference matrix

        diff_matrix = np.zeros((len(calculated_shifts), len(exp_peaks)))

        for ind1, i in enumerate(scaled_shifts):
            for ind2, j in enumerate(exp_peaks):
                diff_matrix[ind1, ind2] = j - i

        ####find any errors larger than 10 ppm and nans
        ####calculate pos matirx

        pos_matrix = carbon_probabilities(diff_matrix, scaled_mu, scaled_std)

        pos_matrix[abs(diff_matrix) >= 10] = 0

        pos_matrix[np.isnan(pos_matrix)] = 0

        ####calculate amp matrix

        amp_matrix = amp_kde(
            total_spectral_ydata, picked_peaks, pos_matrix, calculated_shifts
        )

        ####duplicate the pos matrix along the horizontal to allow multiple assignment weighting

        pos_matrixc = copy.copy(pos_matrix)

        for d in range(0, len(calculated_shifts) - 1):
            pos_matrix = np.hstack((pos_matrix, pos_matrixc))

        ####calculate the probability matrix

        prob_matrix = (pos_matrix * amp_matrix) ** 0.5

        ####check for any shifts that have zero probabilites for all peaks

        b = np.where(np.sum(prob_matrix, 1) == 0)

        prob_matrix = np.delete(prob_matrix, b, 0)

        unassignable_shifts = calculated_shifts[b]

        copy_calc_shifts = np.delete(copy_calc_shifts, b)

        copy_labels = np.delete(C_labels, b)

        ####do the assignment

        vertical_ind, horizontal_ind = optimise(1 - prob_matrix)

        horizontal_ind = horizontal_ind % len(picked_peaks)

        opt_peaks = exp_peaks[horizontal_ind]

        opt_shifts = copy_calc_shifts[vertical_ind]

        opt_labels = copy_labels[vertical_ind]

        ####do some sorting

        so = np.argsort(opt_shifts)

        new_assigned_peaks = opt_peaks[so]

        new_assigned_shifts = opt_shifts[so]

        new_assigned_labels = opt_labels[so]

    ############################
    # in the third round only reassign shifts that have had a change of bias

    old_assigned_shifts = copy.copy(new_assigned_shifts)

    old_assigned_peaks = copy.copy(new_assigned_peaks)

    new_assigned_shifts = copy.copy(new_assigned_shifts)

    new_assigned_peaks = copy.copy(new_assigned_peaks)

    new_assigned_labels = copy.copy(new_assigned_labels)

    bias_weights = []

    # find unassigned peaks

    ampdivide = np.zeros(len(picked_peaks))

    peak_amps = total_spectral_ydata[picked_peaks]

    reassign_shifts_ind = []

    for i in old_assigned_peaks:
        w = np.where(exp_peaks == i)

        ampdivide[w] += 1

    c = 0

    for shift, peak in zip(old_assigned_shifts, old_assigned_peaks):

        # find where peaks are within 20 ppm window

        w = np.where((exp_peaks < peak + 10) & (exp_peaks > peak - 10))[0]

        if len(w) > 0:

            # find maximum peak height within this window - when taking into account how many times the peak has already been assigned

            # find amplitude of peak given how many times it has been assigned

            assigned_amp = (
                peak_amps[exp_peaks == peak] / ampdivide[exp_peaks == peak]
            )[0]

            # find amplitude of max peak in the 20 ppm window given how many times it would be assigned if the current shift was assigned to it as well

            div_amps = peak_amps / (ampdivide + 1)

            pi = np.where(exp_peaks == peak)

            div_amps[pi] = peak_amps[pi] / ampdivide[pi]

            max_window_amp = np.max(div_amps[w])

            ratio = max_window_amp / assigned_amp

            if ratio > 1:
                bias_weights.append(ratio)

                reassign_shifts_ind.append(c)

        c += 1

    ####reassign the shifts with a bias above zero in order of bias to peak within ten ppm with largest unassigned amplitude

    bias_weights = np.array(bias_weights)

    reassign_shifts = np.array(old_assigned_shifts)[reassign_shifts_ind]

    s = np.argsort(bias_weights)

    reassign_shifts = reassign_shifts[s]

    reassign_shifts_ind = np.array(reassign_shifts_ind)[s]

    for shift, ind in zip(reassign_shifts, reassign_shifts_ind):

        # find peak this shift is assigned to

        p = new_assigned_peaks[ind]

        pi = np.where(exp_peaks == p)

        new_peak_amps = peak_amps / (ampdivide + 1)

        new_peak_amps[pi] = peak_amps[pi] / (ampdivide[pi])

        # find peaks within 10 ppm

        w = np.where((exp_peaks < p + 10) & (exp_peaks > p - 10))[0]

        if len(w) > 0:
            assigned_peak = exp_peaks[w[np.argmax(new_peak_amps[w])]]

            new_assigned_peaks[ind] = assigned_peak

        # recalculate estimated peak heights

        ampdivide = np.zeros(len(picked_peaks))

        for i in new_assigned_peaks:
            w = np.where(exp_peaks == i)

            ampdivide[w] += 1

    #############################

    # remove cross assignments

    new_assigned_shifts, new_assigned_peaks, new_assigned_labels = (
        removecrossassignments(
            new_assigned_peaks, new_assigned_shifts, new_assigned_labels
        )
    )

    #### sortoutput wrt original H labels

    assigned_labels = []

    assigned_shifts = []

    assigned_peaks = []

    for label in original_C_labels:

        wh = np.where(new_assigned_labels == label)[0]

        assigned_labels.append(label)

        if len(wh) > 0:

            assigned_shifts.append(new_assigned_shifts[wh[0]])

            assigned_peaks.append(new_assigned_peaks[wh[0]])

        else:
            assigned_shifts.append("")

            assigned_peaks.append("")

    return assigned_shifts, assigned_peaks, assigned_labels, scaled_shifts


def external_scale_carbon_shifts(calculated_shifts):
    scaled = calculated_shifts * 0.9601578792266342 - 1.2625604390657088

    return scaled


def internal_scale_carbon_shifts(assigned_shifts, assigned_peaks, calculated_shifts):
    slope, intercept, r_value, p_value, std_err = linregress(
        assigned_shifts, assigned_peaks
    )

    scaled_shifts = calculated_shifts * slope + intercept

    return scaled_shifts, slope, intercept


def amp_weighting(
    total_spectral_ydata, picked_peaks, prob_matrix, shifts, steep_weights
):
    peak_amps = total_spectral_ydata[picked_peaks]

    peak_amps = peak_amps

    peak_amps = peak_amps / np.max(peak_amps)

    duplicated_amps = copy.copy(peak_amps)

    for d in range(0, len(shifts) - 1):
        duplicated_amps = np.hstack((duplicated_amps, peak_amps * (0.25) ** (d + 1)))

    samps = np.sort(peak_amps)

    thresh = samps[max([len(samps) - len(shifts), 0])]

    def weight_curve(x, thresh, steep):

        y = 1 / (1 + np.exp(-steep * (x - thresh)))

        return y

    steep = np.std(peak_amps)

    x = np.linspace(0, 1, 1000)

    plt.plot(x, weight_curve(x, thresh, 100 * steep))
    plt.plot(x, weight_curve(x, thresh, 400 * steep))

    amp_matrix = np.zeros((len(shifts), len(duplicated_amps)))

    for i in range(0, len(amp_matrix[:, 0])):
        amp_matrix[i, :] = weight_curve(
            duplicated_amps, thresh, steep * 100 ** steep_weights[i]
        )

    plt.plot(peak_amps, weight_curve(peak_amps, thresh, 100 * steep), "o", color="C0")
    plt.plot(peak_amps, weight_curve(peak_amps, thresh, 400 * steep), "co", color="C1")
    plt.show()

    return amp_matrix


def amp_kde(total_spectral_ydata, picked_peaks, prob_matrix, shifts):
    peak_amps = total_spectral_ydata[picked_peaks]

    peak_amps = peak_amps

    peak_amps = peak_amps / np.max(peak_amps)

    x = np.linspace(0, 1, 1000)

    kde = gaussian_kde(peak_amps)

    y = kde.evaluate(x)

    ddy = np.diff(y, 2)

    ddy1 = np.roll(ddy, 1)

    ddyn1 = np.roll(ddy, -1)

    w = np.where((ddy[1:-1] > ddy1[1:-1]) & (ddy[1:-1] > ddyn1[1:-1]))[0] + 2

    # add zero and one values

    if w[0] != 0:
        w = np.hstack((0, w))

    if w[-1] != len(ddy) - 2:
        w = np.hstack((w, len(y) - 1))

    minima = x[w]

    i = 0

    groups = np.zeros(len(peak_amps))

    number_in_group = []

    for m, m1 in zip(minima[:-1], minima[1:]):
        w = np.where((peak_amps > m) & (peak_amps <= m1))[0]

        groups[w] = i

        number_in_group.append(len(w))

        i += 1

    groups = groups.astype(int)

    # convert group numbers to weights

    cumsum = np.cumsum(number_in_group[::-1])[::-1]

    weight_values = len(shifts) / cumsum

    weight_values /= np.max(weight_values)

    peak_weights = weight_values[groups]

    duplicated_weights = copy.copy(peak_weights)

    # do multiple assignment weights

    for d in range(0, len(shifts) - 1):
        duplicated_weights = np.hstack(
            (
                duplicated_weights,
                peak_weights * (0.125 ** (np.max(groups) - groups + 1)) ** (d + 1),
            )
        )

    duplicated_weightsc = copy.copy(duplicated_weights)

    # duplicate along vertical

    for d in range(0, len(shifts) - 1):
        duplicated_weights = np.vstack((duplicated_weights, duplicated_weightsc))

    # renormalise

    for i in range(duplicated_weights.shape[0]):
        duplicated_weights[i, :] = duplicated_weights[i, :] / np.sum(
            duplicated_weights[i, :]
        )

    return duplicated_weights


def multiple_assignment_weighting(prob_matrix):
    # shifts are columns
    # peaks are rows

    pmcopy = copy.copy(prob_matrix)

    for i, shift in enumerate(pmcopy[:, 0]):
        prob_matrix = np.hstack((prob_matrix, pmcopy * (1 / (i + 1))))

    return prob_matrix


def carbon_probabilities(diff_matrix, scaled_mu, scaled_std):

    prob_matrix = norm.pdf(diff_matrix, scaled_mu, scaled_std) / norm.pdf(
        scaled_mu, scaled_mu, scaled_std
    )

    for i in range(prob_matrix.shape[0]):
        prob_matrix[i, :] = prob_matrix[i, :] / np.sum(prob_matrix[i, :])

    return prob_matrix


def simulate_spectrum(spectral_xdata_ppm, calc_shifts):
    y = np.zeros(len(spectral_xdata_ppm))

    for shift in calc_shifts:
        y += lorentzian(spectral_xdata_ppm, 0.001, shift, 0.2)

    return y


def simulate_spectrum(spectral_xdata_ppm, calc_shifts, assigned_peaks, set_exp):
    for ind, shift in enumerate(calc_shifts):
        exp_p = assigned_peaks[ind]

        ind2 = set_exp.index(exp_p)

        y = lorentzian(spectral_xdata_ppm, 0.001, shift, 0.2)

        plt.plot(spectral_xdata_ppm, y + 1.05, color="C" + str(ind2))


def simulate_calc_data(spectral_xdata_ppm, calculated_locations, simulated_ydata):
    ###simulate calcutated data

    simulated_calc_ydata = np.zeros(len(spectral_xdata_ppm))

    for peak in calculated_locations:
        y = np.exp(-0.5 * ((spectral_xdata_ppm - peak) / 0.002) ** 2)
        simulated_calc_ydata += y

    scaling_factor = np.amax(simulated_ydata) / np.amax(simulated_calc_ydata)

    simulated_calc_ydata = simulated_calc_ydata * scaling_factor

    return simulated_calc_ydata


def lorentzian(p, w, p0, A):
    x = (p0 - p) / (w / 2)
    L = A / (1 + x**2)

    return L


def remove_iodine(sdffile, lbls, shifts):
    f = sdffile + ".sdf"

    mol = Chem.MolFromMolFile(f, removeHs=False)

    CI = []

    for atom in mol.GetAtoms():
        if atom.GetSymbol() == "C":
            for neighbor in atom.GetNeighbors():
                if neighbor.GetSymbol() == "I":
                    CI.append(f"C{atom.GetIdx()+1}")

    # remove these carbons

    for C in CI:

        ind = lbls.index(C)

        lbls.remove(C)

        for l in shifts:
            l.pop(ind)

    return lbls, shifts


def removecrossassignments(exp, calc, labels):
    # sort these in decending order

    s = np.argsort(calc)[::-1]

    calc = calc[s]

    exp = exp[s]

    labels = labels[s]

    # generate difference matrix
    switch = 0

    expcopy = np.array(exp)

    while switch == 0:

        swapm = np.zeros([len(calc), len(calc)])

        for i, Hi in enumerate(expcopy):
            for j, Hj in enumerate(expcopy):

                if i > j:

                    swapm[i, j] = 0
                else:
                    swapm[i, j] = round(Hi - Hj, 1)

        w = np.argwhere(swapm < 0)

        if len(w > 0):
            expcopy[w[0]] = expcopy[w[0][::-1]]

        else:
            switch = 1

    return calc, expcopy, labels



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 12 13:15:47 2015

@author: ke291

Contains all of the MacroModel specific code for input generation, calculation
execution and output interpretation. Called by PyDP4.py.
"""

import os
from pathlib import Path
import shutil
import sys
import subprocess
import shutil
import time
import re
import logging

from dp5.conformer_search.base_cs_method import BaseConfSearch, ConfData

logger = logging.getLogger(__name__)


class ConfSearchMethod(BaseConfSearch):

    def __init__(self, settings):
        super().__init__(settings)
        self.energy_cutoff = self.settings["energy_cutoff"]

        if settings["force_field"].lower() == "mmff":
            self.ff_code = 10
        elif settings["force_field"].lower() == "opls":
            self.ff_code = 14
        else:
            logger.critical("Invalid force field provided")
            raise ValueError("Invalid force field provided")

        if settings["executable"]["schrodinger"]:
            self.executable = settings["executable"]["schrodinger"]
        else:
            schrodinger_path = os.getenv("SCHRODINGER")
            if schrodinger_path is None:
                if os.path.exists("/usr/local/shared/schrodinger/current"):
                    self.executable = "/usr/local/shared/schrodinger/current"
                else:
                    self.executable = ""
                    logger.error("Could not find Schrodinger folder")
                    logger.error("Please provide the path in your config file")
            else:
                self.executable = schrodinger_path

    def __repr__(self):
        return "MacroModel"

    def prepare_input(self, inputs):
        if os.name == "nt":
            convinp = '"' + self.executable + '/utilities/sdconvert" -isd '
        else:
            convinp = self.executable + "/utilities/sdconvert -isd "

        for input in inputs:
            if not os.path.exists(f"{input}.mae"):
                if self.executable:
                    outp = subprocess.check_output(
                        f"{convinp}{input}.sdf -omae {input}.mae", shell=True
                    )
                else:
                    logger.critical("Cannot convert input to .mae format")
                    raise RuntimeError("Schrodinger path not found!")

            with open((Path.cwd() / f"{input}.com").resolve(), "w") as f:
                cmd = self.command.format(
                    input, self.ff_code, self.settings["step_count"]
                )
                f.write(cmd)
                logger.info(f"Prepared Macromodel input for {input}")

        logger.info("Macromodel input prepared successfully")

        return inputs

    def _run(self):

        completed = 0
        total = len(self.inputs)
        outputs = []

        installed = True
        path_to_exe = os.path.join(self.executable, "bmin")

        if shutil.which(path_to_exe) is None:
            logger.error(f"Could not find MacroModel executable at {path_to_exe}")
            installed = False

        for input in self.inputs:

            # first, check if we have run this already
            if os.path.exists(f"{input}.log"):
                if self.IsMMCompleted(f"{input}.log"):
                    logger.info(f"Valid {input}.log exists")
                    completed += 1
                    outputs.append(f"{input}.log")
                else:
                    logger.warning(f"{input}.log is incomplete, deleting...")
                    os.remove(f"{input}.log")

            else:
                if not installed:
                    logger.critical("Cannot launch conformational search")
                    raise RuntimeError("Path to Schrodinger not found!")

                logger.info(f"Starting conformational search for {input}")
                outp = subprocess.check_output(f"{path_to_exe} {input}", shell=True)

                time.sleep(60)
                while not self.IsMMCompleted(input + ".log"):
                    time.sleep(30)

                completed = completed + 1
                outputs.append(input)

            logger.info(f"{completed} out of {total} inputs processed")

        return outputs

    def _parse_output(self, file):
        output = f"{file}.log"
        if self.IsMMCompleted(output):
            result = self.read_macromodel(file, self.energy_cutoff)
            logger.debug(f"Reading {file}-out.mae output")
            conf_data = ConfData(*result)
        else:
            logger.critical(f"{output} is not completed")
            raise FileNotFoundError(f"{output} containes incomplete calculations")
        return conf_data

    def read_macromodel(self, file, cutoff_energy):

        atoms_to_symbols = {
            "1": "C",
            "2": "C",
            "3": "C",
            "4": "C",
            "5": "C",
            "6": "C",
            "7": "C",
            "8": "C",
            "9": "C",
            "10": "C",
            "11": "C",
            "12": "C",
            "13": "C",
            "14": "O",
            "15": "O",
            "16": "O",
            "18": "O",
            "20": "O",
            "21": "O",
            "23": "O",
            "24": "N",
            "25": "N",
            "26": "N",
            "31": "N",
            "32": "N",
            "38": "N",
            "39": "N",
            "40": "N",
            "41": "H",
            "42": "H",
            "43": "H",
            "44": "H",
            "45": "H",
            "48": "H",
            "49": "S",
            "51": "S",
            "52": "S",
            "53": "P",
            "54": "B",
            "55": "B",
            "56": "F",
            "57": "Cl",
            "58": "Br",
            "59": "I",
            "60": "Si",
            "100": "S",
            "101": "S",
            "102": "Cl",
            "103": "B",
            "104": "F",
            "109": "S",
            "110": "S",
            "113": "S",
            "114": "S",
        }
        readline = re.compile(r'".*?"|\S+')
        atoms = []
        coordinates = []
        charge = None
        energies = []

        with open(f"{file}-out.mae") as f:
            partial_block = False
            while True:
                try:
                    line = next(f)
                    line = line.strip()
                    # get energies
                    if line.startswith("p_m_ct") or line.startswith("f_m_ct"):
                        if line.startswith("p_m_ct"):
                            partial_block = True
                            logger.debug("Reading partial connection table block")
                        else:
                            partial_block = False
                            logger.debug("Reading full connection table block")
                        energy_offset = 0
                        while not ("mmod_Potential_Energy" in line):
                            line = next(f)
                            energy_offset += 1
                        while not ":::" in line:
                            line = next(f)

                        for _ in range(energy_offset):
                            line = next(f)
                        energies.append(float(line))

                    if line.startswith("m_atom["):
                        atype, xpos, ypos, zpos, cpos = None, None, None, None, None
                        offset = -1
                        while not ":::" in line:
                            line = line.strip()
                            if line.startswith("i_m_mmod_type"):
                                atype = offset
                            elif line.startswith("r_m_x_coord"):
                                xpos = offset
                            elif line.startswith("r_m_y_coord"):
                                ypos = offset
                            elif line.startswith("r_m_z_coord"):
                                zpos = offset
                            elif line.startswith("r_m_charge1"):
                                cpos = offset
                            offset += 1
                            line = next(f)

                        line = next(f)
                        _atoms = []
                        coords = []
                        _charge = 0
                        while not ":::" in line:
                            line = readline.findall(line)
                            if atype is not None:
                                _atoms.append(line[atype])
                            coords.append(
                                [float(i) for i in (line[xpos], line[ypos], line[zpos])]
                            )
                            if cpos is not None:
                                _charge += float(line[cpos])
                            line = next(f)
                        if not atoms:
                            atoms = [atoms_to_symbols[i] for i in _atoms]
                        if charge is None:
                            charge = int(_charge)
                        coordinates.append(coords)

                except StopIteration:
                    logger.debug(f"Finished reading {file}-out.mae")
                    break

        final_coordinates = []
        final_energies = []
        e_min = min(energies)
        for coord, energy in zip(coordinates, energies):
            if energy <= e_min + cutoff_energy:
                final_coordinates.append(coord)
                final_energies.append(energy)

        return atoms, final_coordinates, charge, final_energies

    def IsMMCompleted(self, f):
        with open(f, "r") as Gfile:
            outp = Gfile.readlines()

        if os.name == "nt":
            i = -2
        else:
            i = -3

        if "normal termination" in outp[i]:
            return True
        else:
            return False

    command = """{0}.mae
{0}-out.mae
 MMOD       0      1      0      0     0.0000     0.0000     0.0000     0.0000
 FFLD  {1:>6}      1      0      0     1.0000     0.0000     0.0000     0.0000
 BDCO       0      0      0      0    41.5692 99999.0000     0.0000     0.0000
 READ       0      0      0      0     0.0000     0.0000     0.0000     0.0000
 CRMS       0      0      0      0     0.0000     0.2500     0.0000     0.0000
 LMCS  {2:>6}      0      0      0     0.0000     0.0000     3.0000     6.0000
 NANT       0      0      0      0     0.0000     0.0000     0.0000     0.0000
 MCNV       0      0      0      0     0.0000     0.0000     0.0000     0.0000
 MCSS       2      0      0      0    21.0000     0.0000     0.0000     0.0000
 MCOP       1      0      0      0     0.5000     0.0000     0.0000     0.0000
 DEMX       0 333333      0      0    21.0000    42.0000     0.0000     0.0000
 MSYM       0      0      0      0     0.0000     0.0000     0.0000     0.0000
 AUOP       0      0      0      0  2500.0000     0.0000     0.0000     0.0000
 AUTO       0      2      1      1     0.0000    -1.0000     0.0000     2.0000
 CONV       2      0      0      0     0.0010     0.0000     0.0000     0.0000
 MINI       1      0 999999      0     0.0000     0.0000     0.0000     0.0000
"""



import logging
import os
import shutil

from dp5.dft import BaseDFTMethod


logger = logging.getLogger(__name__)


class DFTMethod(BaseDFTMethod):
    def __init__(self, settings):
        super().__init__(settings)

        self.tag = "o"
        self.input_format = ".inp"
        self.output_format = ".out"

        self.executable = self.settings["executable"]["orca"]

        if shutil.which(self.executable) is None:
            logger.warning(
                f"Gaussian executable not found at {self.executable}. Calculations will not run!"
            )

        for i in "noe":
            self.settings[f"{i}_basis_set"] = (
                "def2-tzvp"
                if self.settings[f"{i}_basis_set"] == "def2tzvp"
                else self.settings[f"{i}_basis_set"]
            )

    def __repr__(self) -> str:
        return "ORCA"

    def prepare_command(self, file):
        return super().prepare_command(file)

    def write_file(self, filename, coordinates, atoms, charge, calc_type):

        with open(f"{filename}.inp", "w") as f:

            if self.settings["num_processors"] > 1:
                f.write(f"%pal nprocs {self.settings['num_processors']}\n  end\n")

            # add solvents!!!

            if calc_type == "opt":
                f.write(self._opt_settings())
            elif calc_type == "nmr":
                f.write(self._nmr_settings())
            else:
                f.write(self._energy_settings())

            if self.settings["solvent"]:
                f.write(
                    f"%cpcm\n  smd true\n"
                    f'  smdSolvent "{self.settings["solvent"]}"\n'
                    "end\n"
                )

            f.write(f"* xyz {charge} 1\n")
            for atom, (x, y, z) in zip(atoms, coordinates):
                f.write(f"{atom}  {x:.6f}  {y:.6f}  {z:.6f}\n")
            f.write("*\n")

            if calc_type == "nmr":
                f.write(
                    "%eprnmr\n"
                    "  Ori GIAO\n"
                    "  NMRShielding 2\n"
                    "  Nuclei = all {shift}\n"
                    "end\n"
                )

    functional_dict = {
        "b3lyp": "functional b3lyp_g",
        "m062x": "functional m062x",
        "mpw1pw91": "exchange gga_x_mpw91\n  correlation gga_c_pw91",
    }
    basis_dict = {"def2tzvp": "def2-tzvp"}

    def _opt_settings(self):
        route = (
            "%method\n"
            "  Method dft\n"
            "  RunTyp opt\n"
            f"  {self.functional_dict[self.settings['o_functional'].lower()]}\n"
            "end\n"
            "%basis\n"
            f"  basis \"{self.settings['o_basis_set']}\"\n"
            "end\n"
            "%geom\n"
            f"  MaxIter {self.settings['max_opt_cycles']}\n"
            f"  MaxStep {self.settings['opt_step_size']}\n"
            f"  calc_hess {self.settings['calc_force_constants']}\n"
            "end\n"
        )
        return route

    def _nmr_settings(self):
        route = (
            "%method\n"
            "  Method dft\n"
            "  RunTyp energy\n"
            f"  {self.functional_dict[self.settings['n_functional'].lower()]}\n"
            "end\n"
            "%basis\n"
            f"  basis \"{self.settings['n_basis_set']}\"\n"
            "end\n"
        )
        return route

    def _energy_settings(self):
        route = (
            "%method\n"
            "  Method dft\n"
            "  RunTyp energy\n"
            f"  {self.functional_dict[self.settings['e_functional'].lower()]}\n"
            "end\n"
            "%basis\n"
            f"  basis \"{self.settings['e_basis_set']}\"\n"
            "end\n"
        )
        return route

    def read_file(self, file):
        atoms = []
        coordinates = []
        energy = None
        shieldings = []
        _shieldings = []
        shielding_labels = []
        # check for normal termination
        completed = False
        # check for optimisation convergence
        converged = False
        with open(file, "r") as f:
            while not completed:
                try:
                    line = next(f).strip()  # get new line using next
                    if line == "CARTESIAN COORDINATES (ANGSTROEM)":
                        _atoms = []
                        coordinates = []
                        for _ in range(2):
                            line = next(f).strip()
                        while line:
                            atom, *coords = line.split()
                            _atoms.append(atom)
                            coordinates.append([float(i) for i in coords])
                            line = next(f).strip()
                        if not atoms:
                            atoms = _atoms

                    if line == "CHEMICAL SHIELDING SUMMARY (ppm)":
                        for _ in range(6):
                            line = next(f).strip()
                        while line:
                            num, elem, shielding, _ = line.split()
                            _shieldings.append(
                                (
                                    int(num) + 1,
                                    float(shielding),
                                    "%s%i" % (elem, int(num) + 1),
                                )
                            )

                            line = next(f).strip()

                    if line == "***        THE OPTIMIZATION HAS CONVERGED     ***":
                        converged = True

                    if line.startswith("FINAL SINGLE POINT ENERGY"):
                        *_, _energy = line.split()
                        energy = self.hartree_to_kJ * float(_energy)

                    if line == "****ORCA TERMINATED NORMALLY****":
                        completed = True

                except StopIteration:
                    logger.error(
                        f"Calculations in {file} have not terminated normally."
                    )
                    break
            # write your functions
        # ORCA does not sort by atom :(
        if _shieldings:
            _shieldings.sort(key=lambda x: x[0])
            _, shieldings, shielding_labels = [list(x) for x in zip(*_shieldings)]
        return (
            atoms,
            coordinates,
            energy,
            shieldings,
            shielding_labels,
            completed,
            converged,
        )



# -*- coding: utf-8 -*-
"""
Created on Mon Jan  12 14:42:47 2015

@author: ke291

Takes care of all the NMR description interpretation, equivalent atom
averaging, Boltzmann averaging and DP4 input preparation and running DP4.py. Called by PyDP4.py

FUNCTIONS AFTER REWRITE:
Calculation of NMR shifts based on TMS reference
Equivalent atom averaging
NMR description parsing
NMR raw data interpretation top level organization
"""

import re
import os
import math
import copy
import pickle
import logging

from dp5.nmr_processing import process_carbon, process_proton

from pathlib import Path
import shutil

gasConstant = 8.3145
temperature = 298.15
hartreeEnergy = 2625.499629554010

logger = logging.getLogger(__name__)

# Data structure for loading and keeping all of experimental NMR data in one place.


class NMRData:
    def __init__(self, structure, nmr_file, solvent, output_folder):

        self.cwd = Path(os.getcwd())
        self.structures = structure
        self.InputPath = nmr_file  # Initial structure input file
        self.Type = (
            "desc"  # desc or fid, depending on whether the description or raw data used
        )
        self.OutputFolder = output_folder
        self.Atoms = []  # Element labels
        self.Cshifts = []  # Experimental C NMR shifts
        self.Clabels = []  # Experimental C NMR labels, if any
        self.Hshifts = []  # Experimental H NMR shifts
        self.Hlabels = []  # Experimental H NMR labels, if any
        self.Equivalents = (
            []
        )  # Atoms assumed to be NMR equivalent in computational data
        self.Omits = []
        self.protondata = {}
        self.carbondata = {}

        logger.info(f"Reading NMR data from{self.InputPath}")

        if len(self.InputPath) == 0:
            logger.critical("No NMR Data Added, quitting...")
            quit()

        else:

            for ind1, p in enumerate(self.InputPath):

                if p.exists():

                    if p.is_dir():

                        self.Type = "fid"

                        if p.parts[-1] == "Proton" or p.parts[-1] == "proton":

                            self.ProcessProton(settings, ind1)

                        elif p.parts[-1] == "Carbon" or p.parts[-1] == "carbon":

                            self.ProcessCarbon(settings, ind1)

                    elif p.parts[-1] == "Proton.dx" or p.parts[-1] == "proton.dx":

                        self.Type = "jcamp"

                        self.ProcessProton(settings, ind1)

                    elif p.parts[-1] == "Carbon.dx" or p.parts[-1] == "carbon.dx":

                        self.Type = "jcamp"

                        self.ProcessCarbon(settings, ind1)
                    else:

                        self.Type = "desc"
                        self.ExpNMRFromDesc()

                else:
                    logger.critical("NMR data path does not exist, quitting...")
                    quit()

    def ExpNMRFromDesc(self):

        print("Loading NMR data from " + str(self.InputPath))

        # Reads the experimental NMR data from the file
        ExpNMR_file = open(self.InputPath[0], "r")
        Cexp = ExpNMR_file.readline()
        ExpNMR_file.readline()
        Hexp = ExpNMR_file.readline()

        # Check if exp NMR file contains info about equivalent atoms and read it
        # into an array
        # Also reads a list of atoms to omit from analysis

        equivalents = []
        omits = []

        ExpNMR_file.readline()
        for line in ExpNMR_file:
            if not "OMIT" in line and len(line) > 1:
                equivalents.append(line[:-1].split(","))
            elif "OMIT" in line:
                omits.extend(line[5:-1].split(","))

        ExpNMR_file.close()

        self.Clabels, self.C_exp = self.ParseExp(Cexp)
        self.Hlabels, self.H_exp = self.ParseExp(Hexp)
        self.Equivalents = equivalents
        self.Omits = omits

    def ParseExp(self, exp):

        if len(exp) > 0:
            # Replace all 'or' and 'OR' with ',', remove all spaces and 'any'
            texp = re.sub(r"or|OR", ",", exp, flags=re.DOTALL)
            texp = re.sub(r" ", "", texp, flags=re.DOTALL)

            # Get all assignments, split mulitassignments
            expLabels = re.findall(r"(?<=\().*?(?=\)|;)", texp, flags=re.DOTALL)
            expLabels = [x.replace("any", "") for x in expLabels]
            expLabels = [x.split(",") for x in expLabels]

            # Remove assignments and get shifts
            ShiftData = (re.sub(r"\(.*?\)", "", exp.strip(), flags=re.DOTALL)).split(
                ","
            )

            print(ShiftData)

            expShifts = [float(x) for x in ShiftData]

        else:

            expLabels = []
            expShifts = []

        return expLabels, expShifts

    def ProcessProton(self, settings, ind):

        pdir = self.OutputFolder / "Pickles"

        gdir = self.OutputFolder / "Graphs"

        NMR_file = self.InputPath[ind]

        if not Path(gdir).exists():

            os.mkdir(gdir)

            os.mkdir(gdir / self.structures[0])

        else:

            if not Path(gdir / self.structures[0]).exists():

                os.mkdir(gdir / self.structures[0])

        if not pdir.exists():

            os.mkdir(pdir)

            os.mkdir(pdir / self.structures[0])

        else:

            if not Path(pdir / self.structures[0]).exists():

                os.mkdir(pdir / self.structures[0])

        if Path(pdir / self.structures[0] / "protondata").exists():

            self.protondata = pickle.load(
                open(pdir / self.structures[0] / "protondata", "rb")
            )

            self.Hshifts = self.protondata["exppeaks"]

        else:

            protondata = {}

            (
                protondata["exppeaks"],
                protondata["xdata"],
                protondata["ydata"],
                protondata["integrals"],
                protondata["peakregions"],
                protondata["centres"],
                protondata["cummulativevectors"],
                protondata["integralsum"],
                protondata["picked_peaks"],
                protondata["params"],
                protondata["sim_regions"],
            ) = process_proton(NMR_file, settings, self.Type)

            pickle.dump(
                protondata,
                Path(pdir / self.structures[0] / "protondata").open(mode="wb+"),
            )

            self.Hshifts = protondata["exppeaks"]

            self.protondata = protondata

    def ProcessCarbon(self, settings, ind):

        pdir = self.OutputFolder / "Pickles"

        gdir = self.OutputFolder / "Graphs"

        NMR_file = self.InputPath[ind]

        if not Path(gdir).exists():

            os.mkdir(gdir)

            os.mkdir(gdir / self.structures[0])

        else:

            if not Path(gdir / self.structures[0]).exists():

                os.mkdir(gdir / self.structures[0])

        if not pdir.exists():

            os.mkdir(pdir)

            os.mkdir(pdir / self.structures[0])

        else:

            if not Path(pdir / self.structures[0]).exists():

                os.mkdir(pdir / self.structures[0])

        if Path(pdir / self.structures[0] / "carbondata").exists():

            self.carbondata = pickle.load(
                open(pdir / self.structures[0] / "carbondata", "rb")
            )

            self.Cshifts = self.carbondata["exppeaks"]

        else:

            carbondata = {}

            (
                carbondata["ydata"],
                carbondata["xdata"],
                carbondata["corrdistance"],
                carbondata["uc"],
                carbondata["exppeaks"],
                carbondata["simulated_ydata"],
                carbondata["removed"],
            ) = process_carbon(NMR_file, settings, self.Type)

            pickle.dump(
                carbondata,
                Path(pdir / self.structures[0] / "carbondata").open(mode="wb+"),
            )

            # pickle.dump(a, Path("/Users/Maidenhair/Desktop/text.txt").open(mode="wb+"))

            self.carbondata = carbondata
            self.Cshifts = carbondata["exppeaks"]


def NMRDataValid(Isomers):

    for isomer in Isomers:
        if len(isomer.ConformerShieldings) == 0:
            return False

    return True


def CalcNMRShifts(Isomers, settings):

    print(
        "WARNING: NMR shift calculation currently ignores the instruction to exclude atoms from analysis"
    )
    for i, iso in enumerate(Isomers):

        BShieldings = iso.BoltzmannShieldings

        Cvalues = []
        Hvalues = []
        Clabels = []
        Hlabels = []

        for a, atom in enumerate(iso.Atoms):

            if atom == "C":
                shift = (settings.TMS_SC_C13 - BShieldings[a]) / (
                    1 - (settings.TMS_SC_C13 / 10**6)
                )
                Cvalues.append(shift)
                Clabels.append("C" + str(a + 1))

            if atom == "H":
                shift = (settings.TMS_SC_H1 - BShieldings[a]) / (
                    1 - (settings.TMS_SC_H1 / 10**6)
                )
                Hvalues.append(shift)
                Hlabels.append("H" + str(a + 1))

        Isomers[i].Cshifts = Cvalues
        Isomers[i].Hshifts = Hvalues

        Isomers[i].Clabels = Clabels
        Isomers[i].Hlabels = Hlabels

        print("C shifts for isomer " + str(i) + ": ")
        print(", ".join(["{0:.3f}".format(x) for x in Isomers[i].Cshifts]))

        print("H shifts for isomer " + str(i) + ": ")
        print(", ".join(["{0:.3f}".format(x) for x in Isomers[i].Hshifts]))

        for conf in iso.ConformerShieldings:

            Cconfshifts = []
            Hconfshifts = []

            for a, atom in enumerate(iso.Atoms):

                if atom == "C":

                    shift = (settings.TMS_SC_C13 - conf[a]) / (
                        1 - (settings.TMS_SC_C13 / 10**6)
                    )
                    Cconfshifts.append(shift)

                if atom == "H":
                    shift = (settings.TMS_SC_H1 - conf[a]) / (
                        1 - (settings.TMS_SC_H1 / 10**6)
                    )
                    Hconfshifts.append(shift)

            Isomers[i].ConformerCShifts.append(Cconfshifts)
            Isomers[i].ConformerHShifts.append(Hconfshifts)

    return Isomers


def PrintConformationData(AllSigConfs):
    """REDUNDANT"""
    # Make a list of populations and corresponding files for reporting
    # significant conformations
    """from operator import itemgetter
    ConfsPops = [list(x) for x in zip(args, populations)]
    ConfsPops.sort(key=itemgetter(1), reverse=True)
    totpop = 0
    i = 0
    while totpop < 0.8:
        totpop += ConfsPops[i][1]
        i += 1
    SigConfs = ConfsPops[:i]"""
    for Es, pops in zip(RelEs, populations):
        print(
            "\nConformer relative energies (kJ/mol): "
            + ", ".join(["{:5.2f}".format(float(x)) for x in Es])
        )

        print(
            "\nPopulations (%): "
            + ", ".join(["{:4.1f}".format(float(x) * 100) for x in pops])
        )

    for i, SigConfs in enumerate(AllSigConfs):
        print(
            "\nNumber of significant conformers for isomer "
            + str(i + 1)
            + ": "
            + str(len(SigConfs))
            + "\n(pop, filename)"
        )
        for conf in SigConfs:
            print("   " + format(conf[1] * 100, "4.2f") + "%   " + conf[0])
        print("----------------")
        print(
            "   " + format(100 * sum([x[1] for x in SigConfs]), "4.2f") + "%   in total"
        )


def RemoveEquivalents(Noutp, equivs, OldCval, OldHval, OldClabels, OldHlabels):
    """
    Currently unused, will use older version to divine the original purpose
    """
    Cvalues = list(OldCval)
    Hvalues = list(OldHval)
    Clabels = list(OldClabels)
    Hlabels = list(OldHlabels)

    for eqAtoms in equivs:

        eqSums = [0.0] * Noutp
        eqAvgs = [0.0] * Noutp

        if eqAtoms[0][0] == "H":
            # print eqAtoms, Hlabels
            for atom in eqAtoms:
                eqIndex = Hlabels.index(atom)
                for ds in range(0, Noutp):
                    eqSums[ds] = eqSums[ds] + Hvalues[ds][eqIndex]
            for ds in range(0, Noutp):
                eqAvgs[ds] = eqSums[ds] / len(eqAtoms)

            # Place the new average value in the first atom shifts place
            target_index = Hlabels.index(eqAtoms[0])
            for ds in range(0, Noutp):
                Hvalues[ds][target_index] = eqAvgs[ds]

            # Delete the redundant atoms from the computed list
            # start with second atom - e.g. don't delete the original one
            for atom in range(1, len(eqAtoms)):
                del_index = Hlabels.index(eqAtoms[atom])
                del Hlabels[del_index]
                for ds in range(0, Noutp):
                    del Hvalues[ds][del_index]

        if eqAtoms[0][0] == "C":
            for atom in eqAtoms:
                eqIndex = Clabels.index(atom)
                for ds in range(0, Noutp):
                    eqSums[ds] = eqSums[ds] + Cvalues[ds][eqIndex]
            for ds in range(0, Noutp):
                eqAvgs[ds] = eqSums[ds] / len(eqAtoms)

            # Place the new average value in the first atom shifts place
            target_index = Clabels.index(eqAtoms[0])
            for ds in range(0, Noutp):
                Cvalues[ds][target_index] = eqAvgs[ds]

            # Delete the redundant atoms from the computed list
            # start with second atom - e.g. don't delete the original one
            for atom in range(1, len(eqAtoms)):
                del_index = Clabels.index(eqAtoms[atom])
                del Clabels[del_index]
                for ds in range(0, Noutp):
                    del Cvalues[ds][del_index]

    return Cvalues, Hvalues, Clabels, Hlabels


def PairwiseAssignment(Isomers, NMRData):

    # for each isomer sort the experimental and calculated shifts

    for iso in Isomers:

        sortedCCalc = sorted(iso.Cshifts, reverse=True)
        sortedHCalc = sorted(iso.Hshifts, reverse=True)

        sortedCExp = sorted(NMRData.Cshifts, reverse=True)
        sortedHExp = sorted(NMRData.Hshifts, reverse=True)

        assignedCExp = [""] * len(sortedCCalc)
        assignedHExp = [""] * len(sortedHCalc)

        tempCCalcs = list(iso.Cshifts)
        tempHCalcs = list(iso.Hshifts)

        # do the assignment in order of chemical shift starting with the largest

        # Carbon

        for exp, shift in zip(sortedCExp, sortedCCalc):

            ind = tempCCalcs.index(shift)

            assignedCExp[ind] = exp

            tempCCalcs[ind] = ""

        # Proton

        for exp, shift in zip(sortedHExp, sortedHCalc):

            ind = tempHCalcs.index(shift)

            assignedHExp[ind] = exp

            tempHCalcs[ind] = ""

        # update isomers class

        iso.Cexp = assignedCExp
        iso.Hexp = assignedHExp

    return Isomers



import numpy as np
from random import shuffle

from tensorflow.keras.utils import Sequence


class GraphSequence(Sequence):
    
    def __init__(self, inputs, y=None, batch_size=1, shuffle=True,
                 final_batch=True):
        """ A keras.Sequence generator to be passed to model.fit_generator. (or
        any other *_generator method.) Returns (inputs, y) tuples where
        molecule feature matrices have been stitched together. Offsets the
        connectivity matrices such that atoms are indexed appropriately.

        batch_size: number of molecules per batch
        shuffle : whether to shuffle the input data
        final_batch : whether to include the final, incomplete batch

        """
        self._inputs = inputs
        self._y = np.asarray(y) if y is not None else None
        self._input_keys = list(inputs[0].keys())
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.final_batch = final_batch
        

    def __len__(self):
        """ Total number of batches """
        if self.final_batch:
            return int(np.ceil(len(self._inputs) / float(self.batch_size)))
        else:
            return int(np.floor(len(self._inputs) / float(self.batch_size)))

    def on_epoch_end(self):
        if self.shuffle:
            indices = np.arange(0, len(self._inputs))
            np.random.shuffle(indices)
            self._inputs = [self._inputs[i] for i in indices]
            if self._y is not None:
                self._y = self._y[indices]
    
    def __getitem__(self, idx):
        """ Calculate the feature matrices for a whole batch (with index `i` <
        self.__len__). This involves adding offsets to the indices for each
        atom in the connectivity matrix; such that atoms and bonds in later
        molecules still refer to the correct atoms.

        """
        batch_indexes = idx * self.batch_size + np.arange(0, self.batch_size)
        batch_indexes = batch_indexes[batch_indexes < len(self._inputs)]

        batch_data = {
            key: self._concat([self._inputs[i][key] for i in batch_indexes])
            for key in self._input_keys}
    
        # Offset the connectivity matrix to account for the multiple graphs per
        # batch
        offset = _compute_stacked_offsets(
            batch_data['n_atom'], batch_data['n_bond'])

        batch_data['connectivity'] += offset[:, np.newaxis]
        
        # Compute graph indices with shape (n_atom,) that indicate to which
        # molecule each atom belongs.
        n_graphs = len(batch_indexes)
        # batch_data['node_graph_indices'] = np.repeat(
        #     np.arange(n_graphs), batch_data['n_atom'])

        batch_data = self.process_data(batch_data)

        # Keras takes to options, one (x, y) pairs, or just (x,) pairs if we're
        # doing predictions. Here, if we've specified a y matrix, we return the
        # x,y pairs for training, otherwise just return the x data.
        if self._y is not None:
            return (batch_data, np.concatenate(self._y[batch_indexes]).reshape(-1,1))

        else:
            return batch_data
        
    def process_data(self, batch_data):
        """ function to add additional processing to batch data before returning """

        # These aren't used currently, so I pop them. But we might need them at
        # a later time.
        del batch_data['n_atom']
        del batch_data['n_bond']
        
        return batch_data

    
    def _concat(self, to_stack):
        """ function to stack (or concatentate) depending on dimensions """

        if np.asarray(to_stack[0]).ndim >= 2:
            return np.concatenate(to_stack)
        
        else:
            return np.hstack(to_stack)


def _compute_stacked_offsets(sizes, repeats):
    """ Computes offsets to add to indices of stacked np arrays.
    When a set of np arrays are stacked, the indices of those from the second on
    must be offset in order to be able to index into the stacked np array. This
    computes those offsets.

    Args:
        sizes: A 1D sequence of np arrays of the sizes per graph.
        repeats: A 1D sequence of np arrays of the number of repeats per graph.
    Returns:
        The index offset per graph.
    """
    return np.repeat(np.cumsum(np.hstack([0, sizes[:-1]])), repeats)



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 17 15:20:18 2014

@author: ke291

Gets called by PyDP4.py if J value analysis based on Karplus equation is used
Based on C. A. G. Haasnoot, F. A. A. M. de Leeuw and C. Altona;
Tetrahedron 36 2783-2792 (1980)
"""

import numpy as np
from math import sqrt, pi, cos, sin, acos, atan2
import sys

try:
    from openbabel.openbabel import OBConversion, OBMol, OBAtomAtomIter, OBMolAtomIter
except ImportError:
    from openbabel import *

#Parameter sets B-E for generalised Karplus equation from
#C. A. G. Haasnoot, F. A. A. M. de Leeuw and C. Altona
#Tetrahedron 36 2783-2792 (1980)
Params = [[13.70,-0.73,0.0,0.56,-2.47,16.9,0.14],
          [13.89,-0.98,0.0,1.02,-3.40,14.9,0.24],
          [13.22,-0.99,0.0,0.87,-2.46,19.9,0.00],
          [13.24,-0.91,0.0,0.53,-2.41,15.5,0.19]]

#Electronegativities up to iodine
ENs = [2.20,0.00,0.98,1.57,2.04,2.55,3.04,3.44,3.98,0.00,
       0.93,1.31,1.61,1.90,2.19,2.58,3.16,0.00,0.82,1.00,
       1.36,1.54,1.63,1.66,1.55,1.83,1.88,1.91,1.90,1.65,
       1.81,2.01,2.18,2.55,2.96,3.00,0.82,0.95,1.22,1.33,
       1.60,2.16,1.90,2.20,2.28,2.20,1.93,1.69,1.78,1.96,
       2.05,2.10,2.66]

#input format - "g09" for gaussian, "nwo" for NWChem
def Karplus(f, inputformat):

    obconversion = OBConversion()
    obconversion.SetInFormat(inputformat)
    obmol = OBMol()

    obconversion.ReadFile(obmol, f)

    obmol.ConnectTheDots()
    obmol.Kekulize()
    
    DihedralHs = []
    for atom in OBMolAtomIter(obmol):
        if atom.GetAtomicNum() == 1:
            DihedNeighbours = GetDihedralHs(atom)
            if DihedNeighbours!=0:
                DihedralHs.append([atom.GetIdx()] + DihedNeighbours)                
    
    if len(DihedralHs)==0:
        print("No dihedral protons found, Karplus J value prediction " + \
            "impossible, quitting.")
        quit()
    
    Jmatrix, Jlabels = CalcJMatrix(obmol, DihedralHs)
    Jlabels = [str(x) for x in Jlabels]
    return Jmatrix, Jlabels


def GetDihedralHs(atom):
    sp3carbonFound = False
    #find the proton neigbour, check that it is sp3 carbon
    for NbrAtom in OBAtomAtomIter(atom):
        if NbrAtom.GetAtomicNum() == 6 and NbrAtom.GetHyb() == 3:
            carbon = NbrAtom
            sp3carbonFound = True
            break
    
    if not sp3carbonFound:
        return 0
    #find the carbon neibours, check that they are sp3 hybridized
    ncarbons = []
    for NbrAtom in OBAtomAtomIter(carbon):
        if NbrAtom.GetAtomicNum() == 6 and NbrAtom.GetHyb() == 3:
            ncarbons.append(NbrAtom)
    #check that the carbon neighbours have protons and return them
    nHs = []
    for ncarbon in ncarbons:
        for NbrAtom in OBAtomAtomIter(ncarbon):
            if NbrAtom.GetAtomicNum() == 1:
                nHs.append(NbrAtom.GetIdx())
    
    if len(nHs) == 0:
        return 0
    else:
        return nHs


def CalcJMatrix(obmol, DihedralHs):
    mat = [[0.0 for x in range(len(DihedralHs))] for y in range(len(DihedralHs))]
    labels = [x[0] for x in DihedralHs]
    
    for x, row in enumerate(DihedralHs):
        atom1 = row[0]
        for atom2 in row[1:]:
            y = labels.index(atom2)
            J = CalcJ(obmol.GetAtom(atom1), obmol.GetAtom(atom2))
            mat[x][y] = J
            mat[y][x] = J
    
    return mat, labels


def CalcJ(atom1, atom2):
    
        #Get the 2 carbon atoms joining the protons
    for NbrAtom in OBAtomAtomIter(atom1):
        if NbrAtom.GetAtomicNum() == 6 and NbrAtom.GetHyb() == 3:
            carbon1 = NbrAtom
            break
    for NbrAtom in OBAtomAtomIter(atom2):
        if NbrAtom.GetAtomicNum() == 6 and NbrAtom.GetHyb() == 3:
            carbon2 = NbrAtom
            break
    #Get the 2 planes formed by the carbons and each of the protons
    normal1, d = FindPlane(carbon1, atom1, carbon2)
    normal2, d = FindPlane(carbon2, atom2, carbon1)
    
    sign = VectAngleSign(BondVect(carbon1, atom1), BondVect(carbon2, atom2),
                         BondVect(carbon1, carbon2))
    angle = VectorAngle(normal1, normal2)
    if dotproduct(BondVect(carbon1, atom1), BondVect(carbon2, atom2)) > 0 and \
        angle < 0.5*pi:
        dihedral = angle*sign
    else:
        dihedral = (pi-angle)*sign
    
    #Count the substituents, get their electronegativities
    nSubst = 0
    relENs = []
    ENcorrs = []
    signs = []
    
    for NbrAtom in OBAtomAtomIter(carbon1):
        ANum = NbrAtom.GetAtomicNum()
        if NbrAtom.GetIdx() != carbon2.GetIdx() and ANum != 1:
            nSubst += 1
            relEN, ENcorr = GetENCorrection(NbrAtom, carbon1)
            relENs.append(relEN)
            ENcorrs.append(ENcorr)
            signs.append(VectAngleSign(BondVect(carbon1, atom1),
                                       BondVect(carbon1, NbrAtom),
                                       BondVect(carbon1, carbon2)))
                        
    for NbrAtom in OBAtomAtomIter(carbon2):
        ANum = NbrAtom.GetAtomicNum()
        if NbrAtom.GetIdx() != carbon1.GetIdx() and ANum != 1:
            nSubst += 1
            relEN, ENcorr = GetENCorrection(NbrAtom, carbon2)
            relENs.append(relEN)
            ENcorrs.append(ENcorr)
            signs.append(VectAngleSign(BondVect(carbon2, atom2),
                                       BondVect(carbon2, NbrAtom),
                                       BondVect(carbon2, carbon1)))
    
    SubstEffects = 0
    if nSubst < 2:
        for relEN, ENcorr, sign in zip(relENs, ENcorrs, signs):
            corrRelEN = relEN - (Params[0][6])*ENcorr
            SubstEffects += SubstEffect(dihedral, corrRelEN, sign, 0)
            
        J = Params[0][0]*cos(dihedral)**2 + Params[0][1]*cos(dihedral)
    else:
        for relEN, sign in zip(relENs, signs):
            corrRelEN = relEN - (Params[nSubst-1][6])*ENcorr
            SubstEffects += SubstEffect(dihedral, corrRelEN, sign, nSubst-1)
            
        J = Params[nSubst-1][0]*cos(dihedral)**2 + Params[nSubst-1][1]*cos(dihedral)
    
    return J + SubstEffects


def GetENCorrection(atom, satom):
    Anum = atom.GetAtomicNum()
    if Anum != 6:
        return ENs[Anum-1]-ENs[0], 0.0
    else:
        ENcorr = 0.0
        for NbrAtom in OBAtomAtomIter(atom):
            if NbrAtom.GetIdx() != satom.GetIdx():
                neighbAnum = NbrAtom.GetAtomicNum()
                ENcorr += ENs[neighbAnum-1]-ENs[0]
        
        return ENs[Anum-1]-ENs[0], ENcorr


def SubstEffect(dihedral, relEN, sign, pset):
    
    return relEN*(Params[pset][3] + Params[pset][4]*
                  cos(sign*dihedral + Deg2Rad(Params[pset][5])*abs(relEN))**2)


def BondVect(atom1, atom2):
    return [atom2.x() - atom1.x(),atom2.y() - atom1.y(),atom2.z() - atom1.z()]


def VectorAngle(v1, v2):
    
    cp = crossproduct(v1, v2)
    return atan2(length(cp),dotproduct(v1, v2))


def VectorAngle2(normal1, normal2):
    
    dotprod = sum((a*b) for a, b in zip(normal1, normal2))
    length1 = sqrt(sum([x**2 for x in normal1]))
    length2 = sqrt(sum([x**2 for x in normal2]))
    
    return acos(dotprod/(length1*length2))


def VectAngleSign(v1, v2, axis):
    
    crossprod = crossproduct(v1, v2)
    return np.sign(dotproduct(crossprod, axis))


def GetUnitVector(Atom1, Atom2):
    vector = []
    vector.append(Atom2.x() - Atom1.x())
    vector.append(Atom2.y() - Atom1.y())
    vector.append(Atom2.z() - Atom1.z())

    length = np.linalg.norm(vector)
    return [x/length for x in vector]


#Given 3 atoms, finds a plane defined by a normal vector and d
def FindPlane(atom1, atom2, atom3):

    vector1 = [atom2.x() - atom1.x(), atom2.y() - atom1.y(),
               atom2.z() - atom1.z()]
    vector2 = [atom3.x() - atom1.x(), atom3.y() - atom1.y(),
               atom3.z() - atom1.z()]
    cross_product = crossproduct(vector1, vector2)
    d = cross_product[0] * atom1.x() - cross_product[1] * atom1.y() + \
        cross_product[2] * atom1.z()
    
    return cross_product, d


def Deg2Rad(x):
    return x*0.0174532925


def crossproduct(v1, v2):
    product = [0, 0, 0]
    product[0] = v1[1]*v2[2]-v1[2]*v2[1]
    product[1] = v1[2]*v2[0]-v1[0]*v2[2]
    product[2] = v1[0]*v2[1]-v1[1]*v2[0]
    return product


def dotproduct(v1, v2):
        return sum((a*b) for a, b in zip(v1, v2))


def length(v):
    return sqrt(dotproduct(v, v))


def angle(v1, v2):
    return acos(dotproduct(v1, v2) / (length(v1) * length(v2)))


from pathlib import Path
import logging
from typing import List, Union, Dict

from rdkit import Chem
from rdkit.Chem import AllChem, EnumerateStereoisomers


logger = logging.getLogger(__name__)


def write_to_sdf(mol: Chem.rdchem.Mol, relative_path: Path):
    """
    Writes rdkit Mol object to a specified path

    arguments:
    - mol: RDKit Mol object
    - relative_path: path to write the file

    returns:
    - input_file: relative path to file from current working directory
    """
    path = Path.cwd() / relative_path
    writer = Chem.SDWriter(str(path))
    writer.write(mol)
    return relative_path


def cleanup_3d(mol):
    """
    Generates a 3D conformer of a molecule.

    arguments:
    - mol: RDKit mol object

    returns:
    - a reasonable conformer
    """
    mol = AllChem.AddHs(mol, addCoords=True)
    cid = AllChem.EmbedMolecule(
        mol,
        randomSeed=0,
        forceTol=0.0135,
    )
    if cid == -1:
        raise ValueError("Molecule could not be sanitised")
    AllChem.MMFFOptimizeMolecule(mol, mmffVariant="MMFF94s")
    Chem.rdmolops.AssignStereochemistryFrom3D(mol)

    return mol


def read_sdf(input_file: str):

    fullf = Path(input_file).resolve()
    mol = Chem.MolFromMolFile(str(fullf), removeHs=False, sanitize=True)
    mol = Chem.AddHs(mol)

    return mol


def read_textfile(input_file: str, input_type: str):
    input_type = input_type.lower()

    input_readers = {
        "smiles": Chem.MolFromSmiles,
        "smarts": Chem.MolFromSmarts,
        "inchi": Chem.MolFromInchi,
    }

    if input_type not in input_readers:
        raise ValueError(f"Cannot parse {input_type}")

    mols = []

    with open(input_file) as f:
        for line in f:
            if line.strip() == "":
                continue
            try:
                mol = input_readers[input_type](line.strip(), sanitize=True)
                mol = Chem.AddHs(mol)
                mols.append(mol)
            except Exception as e:
                # to make an logging.error
                logging.error(f"Error reading line: {line} - {e}")

    return mols


def _generate_diastereomers(
    mol: Chem.rdchem.Mol,
    mutable_atoms: Union[List[int], None] = None,
    double_bonds: bool = False,
) -> List[Chem.rdchem.Mol]:
    """
    Enumerates all diastereomers and double bond isomers.

    Given an rdkit Mol object, generates all possible diastereomers and returns them as a list.
    Note that the function also checks if the resulting molecule can be embedded.

    Parameters:
    - mol (rdkit.Chem.rdchem.Mol): the molecule to have its stereocentres enumerated
    - protected_atoms (list[int], None): the list of atoms with a fixed stereochemistry, numbering starts with 1
    - double_bonds (bool): if double bond configuration is also altered. Defaults to False

    Returns:
    list of Mols
    """
    # copy the input to prevent accidental editing
    target_mol = AllChem.AddHs(mol)

    mutable_atoms = mutable_atoms.copy()

    enum_opts = EnumerateStereoisomers.StereoEnumerationOptions(
        tryEmbedding=False, onlyUnassigned=True, maxIsomers=0, unique=True
    )
    # extract chiral information if present in a molecule
    atom_chiral = [at.GetChiralTag() for at in target_mol.GetAtoms()]
    bond_geometry = [b.GetStereo() for b in target_mol.GetBonds()]

    # Creates list of mutable atoms. If none are specified, retains everything but the first stereocentre.
    atoms_with_stereo = [
        i
        for i, tag in enumerate(atom_chiral, start=1)
        if tag != Chem.rdchem.ChiralType.CHI_UNSPECIFIED
    ]

    if not mutable_atoms:
        if len(atoms_with_stereo) > 0:
            atoms_with_stereo.pop(0)
        mutable_atoms = atoms_with_stereo

    mutable_atoms = [ind - 1 for ind in mutable_atoms]

    # all other atoms will have stereochemistry tag reset
    for atom in target_mol.GetAtoms():
        if atom.GetIdx() in mutable_atoms:
            atom.SetChiralTag(Chem.rdchem.ChiralType.CHI_UNSPECIFIED)

    # scrambles all double bonds: if particular configuration was assigned, it is reset
    if double_bonds == True:
        [
            b.SetStereo(Chem.rdchem.BondStereo.STEREOANY)
            for b, g in zip(target_mol.GetBonds(), bond_geometry)
            if g != Chem.rdchem.BondStereo.STEREONONE
        ]

    result = []
    for isomer in EnumerateStereoisomers.EnumerateStereoisomers(
        target_mol, options=enum_opts
    ):
        isomer.RemoveAllConformers()
        cid = AllChem.EmbedMolecule(isomer, randomSeed=0, forceTol=0.0135)
        if cid >= 0:
            AllChem.MMFFOptimizeMolecule(isomer)
            Chem.rdmolops.AssignStereochemistryFrom3D(isomer)
            result.append(isomer)

    return result


def prepare_inputs(
    input_files: List[str], input_type: str, stereocentres: List[int], workflow: Dict
) -> List[str]:
    """
    Reads files at the path specified by input config, prepares them as required by the user. Returns paths to the new files.

    Arguments:
    - input_file (list[str]): list of relative paths to structure input files. Contains one text file of several SD Files.
    - input_type (str): format of the input file. May be 'sdf', 'smiles', 'inchi', and 'smarts'.
    - stereocentres (list[int]): specifies mutable stereocentres. Defaults to empty list
    - workflow (dict): dictionary of booleans specifying the workflow.

    Returns:
    - mol_paths (list[str]): paths to the transformed files
    """
    # if not sdf, read text
    # generate diastereomers (cleans them) and clean inputs
    # returns paths to inputs
    # in principle, can create list of list of mols, use enumerate
    logger.info(f"Read structures from {input_files}")

    if input_type == "sdf":
        mols = [read_sdf(file) for file in input_files]
        logger.debug("read structures from SD File")
    else:
        mols = read_textfile(input_files[0], input_type)
        logger.debug(f"read structures from {input_type} file")
        input_files = [
            f"{input_type}_mol_{i:03}_.sdf"
            for i, mol in enumerate(range(len(mols)), start=1)
        ]

    if len(mols) < 1:
        raise ValueError("No molecules were provided!")

    logger.info(f"Structures read successfully")

    mols2 = []
    mutable_atoms = stereocentres if len(input_files) == 1 else []

    if workflow["generate"]:
        logger.info("Generating diastereomers")
        mols2 = [_generate_diastereomers(mol, mutable_atoms) for mol in mols]
    elif workflow["cleanup"]:
        logger.info("Generating MMFF geometries for inputs")
        mols2 = [[cleanup_3d(mol)] for mol in mols]
    else:
        if not workflow["conf_search"] and not workflow["dft_opt"]:
            logger.warning(
                "Skipping cleanup without conformational search or DFT optimisation is not recommended!"
            )
        mols2 = [[mol] for mol in mols]

    logger.debug("Preparing to write structure files")
    filenames = []
    for filename, mol in zip(input_files, mols2):
        for i, isomer in enumerate(mol, start=1):
            if len(mol) == 1:
                fname = f"{filename[:-4]}.sdf"
            else:
                fname = f"{filename[:-4]}isomer{i:03}.sdf"
            filenames.append(fname)
            write_to_sdf(isomer, fname)

    return filenames
